"""Protocol engine types to do with positions inside wells."""
from enum import Enum, auto
from typing import Union, Literal

from pydantic import BaseModel, Field


class WellOrigin(str, Enum):
    """Origin of WellLocation offset.

    Props:
        TOP: the top-center of the well
        BOTTOM: the bottom-center of the well
        CENTER: the middle-center of the well
        MENISCUS: the meniscus-center of the well
    """

    TOP = "top"
    BOTTOM = "bottom"
    CENTER = "center"
    MENISCUS = "meniscus"


class PickUpTipWellOrigin(str, Enum):
    """The origin of a PickUpTipWellLocation offset.

    Props:
        TOP: the top-center of the well
        BOTTOM: the bottom-center of the well
        CENTER: the middle-center of the well
    """

    TOP = "top"
    BOTTOM = "bottom"
    CENTER = "center"


class DropTipWellOrigin(str, Enum):
    """The origin of a DropTipWellLocation offset.

    Props:
        TOP: the top-center of the well
        BOTTOM: the bottom-center of the well
        CENTER: the middle-center of the well
        DEFAULT: the default drop-tip location of the well,
            based on pipette configuration and length of the tip.
    """

    TOP = "top"
    BOTTOM = "bottom"
    CENTER = "center"
    DEFAULT = "default"


class WellLocationFunction(int, Enum):
    """The type of well location object to be created."""

    BASE = auto()
    LIQUID_HANDLING = auto()
    PICK_UP_TIP = auto()
    DROP_TIP = auto()


# This is deliberately a separate type from Vec3f to let components default to 0.
class WellOffset(BaseModel):
    """An offset vector in (x, y, z)."""

    x: float = 0
    y: float = 0
    z: float = 0


class WellLocation(BaseModel):
    """A relative location in reference to a well's location."""

    origin: WellOrigin = WellOrigin.TOP
    offset: WellOffset = Field(default_factory=WellOffset)
    volumeOffset: float = Field(
        default=0.0,
        description="""A volume of liquid, in µL, to offset the z-axis offset.""",
    )


class LiquidHandlingWellLocation(BaseModel):
    """A relative location in reference to a well's location.

    To be used with commands that handle liquids.
    """

    origin: WellOrigin = WellOrigin.TOP
    offset: WellOffset = Field(default_factory=WellOffset)
    volumeOffset: Union[float, Literal["operationVolume"]] = Field(
        default=0.0,
        description="""A volume of liquid, in µL, to offset the z-axis offset. When "operationVolume" is specified, this volume is pulled from the command volume parameter.""",
    )


class PickUpTipWellLocation(BaseModel):
    """A relative location in reference to a well's location.

    To be used for picking up tips.
    """

    origin: PickUpTipWellOrigin = PickUpTipWellOrigin.TOP
    offset: WellOffset = Field(default_factory=WellOffset)


class DropTipWellLocation(BaseModel):
    """Like WellLocation, but for dropping tips.

    Unlike a typical WellLocation, the location for a drop tip
    defaults to location based on the tip length rather than the well's top.
    """

    origin: DropTipWellOrigin = DropTipWellOrigin.DEFAULT
    offset: WellOffset = Field(default_factory=WellOffset)


WellLocationType = Union[
    WellLocation,
    LiquidHandlingWellLocation,
    PickUpTipWellLocation,
    DropTipWellLocation,
]



"""opentrons.hardware_control.scripts.repl - cli for hc api

Running this script will create and spin up a hardware controller
and expose it to a python commandline.
"""

import os
from functools import wraps
import asyncio
import logging
from logging.config import dictConfig
from opentrons.hardware_control.api import API
from opentrons.hardware_control.ot3api import OT3API
from opentrons.hardware_control.types import HardwareFeatureFlags

update_firmware = True
has_robot_server = True
if os.environ.get("OPENTRONS_SIMULATION"):
    print("Running with simulators")
    has_robot_server = False
if os.environ.get("OT2", None):
    print(
        '"OT2" env var detected, running with OT2 HC. '
        "If you dont want this, remove the OT2 env var"
    )
    os.environ["OT_API_FF_enableOT3HardwareController"] = "false"
else:
    print("Running with OT3 HC. If you dont want this, set an env var named 'OT2'")
    os.environ["OT_API_FF_enableOT3HardwareController"] = "true"
    if os.environ.get("OT3_DISABLE_FW_UPDATES"):
        update_firmware = False
        print("OT3 firmware updates are disabled")

from code import interact  # noqa: E402
from subprocess import run  # noqa: E402
from typing import Union, Type, Any  # noqa: E402

from opentrons.types import Mount, Point  # noqa: E402
from opentrons.config import feature_flags as ff  # noqa: E402
from opentrons.hardware_control.modules.types import ModuleType  # noqa: E402
from opentrons.hardware_control.types import (  # noqa: E402
    Axis,
    OT3Mount,
    SubSystem,
    GripperProbe,
    CriticalPoint,
)
from opentrons.hardware_control.ot3_calibration import (  # noqa: E402
    calibrate_pipette,
    calibrate_belts,
    delete_belt_calibration_data,
    calibrate_gripper_jaw,
    calibrate_module,
    find_calibration_structure_height,
    find_edge_binary,
    CalibrationMethod,
    find_axis_center,
    gripper_pin_offsets_mean,
)
from opentrons.hardware_control.thread_manager import ThreadManager  # noqa: E402


log = logging.getLogger(__name__)

LOG_CONFIG = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "basic": {"format": "%(asctime)s %(name)s %(levelname)s %(message)s"}
    },
    "handlers": {
        "file_handler": {
            "class": "logging.handlers.RotatingFileHandler",
            "formatter": "basic",
            "filename": "/var/log/repl.log",
            "maxBytes": 5000000,
            "level": logging.INFO,
            "backupCount": 3,
        },
    },
    "loggers": {
        "": {
            "handlers": ["file_handler"],
            "level": logging.INFO,
        },
    },
}

if ff.enable_ot3_hardware_controller():

    HCApi: Union[Type[OT3API], Type[API]] = OT3API

    def build_thread_manager() -> ThreadManager[Union[API, OT3API]]:
        return ThreadManager(
            OT3API.build_hardware_controller,
            use_usb_bus=ff.rear_panel_integration(),
            update_firmware=update_firmware,
            feature_flags=HardwareFeatureFlags.build_from_ff(),
        )

    def wrap_async_util_fn(fn: Any, *bind_args: Any, **bind_kwargs: Any) -> Any:
        @wraps(fn)
        def synchronizer(*args: Any, **kwargs: Any) -> Any:
            return asyncio.new_event_loop().run_until_complete(
                fn(*bind_args, *args, **bind_kwargs, **kwargs)
            )

        return synchronizer

else:

    HCApi = API

    def build_thread_manager() -> ThreadManager[Union[API, OT3API]]:
        return ThreadManager(
            API.build_hardware_controller,
            feature_flags=HardwareFeatureFlags.build_from_ff(),
        )


logging.basicConfig(level=logging.INFO)


def stop_server() -> None:
    run(["systemctl", "stop", "opentrons-robot-server"])


def build_api() -> ThreadManager[Union[API, OT3API]]:
    # NOTE: We are using StreamHandler so when the hw controller is
    # being built we can log firmware update progress to stdout.
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.INFO)
    logging.getLogger().addHandler(stream_handler)
    tm = build_thread_manager()
    logging.getLogger().removeHandler(stream_handler)
    tm.managed_thread_ready_blocking()

    if update_firmware:

        async def _do_update() -> None:
            async for update in tm.update_firmware():
                print(f"Update: {update.subsystem.name}: {update.progress}%")

        asyncio.run(_do_update())

    return tm


def do_interact(api: ThreadManager[Union[API, OT3API]]) -> None:
    interact(
        banner=(
            "Hardware Control API REPL\nCall methods on api like "
            "api.move_to(Mount.RIGHT, Point(400, 400, 500))"
        ),
        local={
            "api": api.sync,
            "Mount": Mount,
            "Point": Point,
            "Axis": Axis,
            "OT3Mount": OT3Mount,
            "SubSystem": SubSystem,
            "GripperProbe": GripperProbe,
            "ModuleType": ModuleType,
            "find_edge": wrap_async_util_fn(find_edge_binary, api),
            "find_calibration_structure_height": wrap_async_util_fn(
                find_calibration_structure_height, api
            ),
            "calibrate_pipette": wrap_async_util_fn(calibrate_pipette, api),
            "calibrate_belts": wrap_async_util_fn(calibrate_belts, api),
            "delete_belt_calibration_data": delete_belt_calibration_data,
            "calibrate_gripper": wrap_async_util_fn(calibrate_gripper_jaw, api),
            "calibrate_module": wrap_async_util_fn(calibrate_module, api),
            "gripper_pin_offsets_mean": gripper_pin_offsets_mean,
            "CalibrationMethod": CalibrationMethod,
            "find_axis_center": wrap_async_util_fn(find_axis_center, api),
            "CriticalPoint": CriticalPoint,
        },
    )


if __name__ == "__main__":
    dictConfig(LOG_CONFIG)
    if has_robot_server:
        stop_server()
    api_tm = build_api()
    do_interact(api_tm)
    api_tm.clean_up()



import asyncio
import contextlib
import logging
from abc import ABC, abstractmethod
from typing import AsyncGenerator, List, Optional
from opentrons.hardware_control.modules.errors import AbsorbanceReaderDisconnectedError
from opentrons_shared_data.errors.exceptions import ModuleCommunicationError


log = logging.getLogger(__name__)


class Reader(ABC):
    @abstractmethod
    async def read(self) -> None:
        """Read some data from an external source."""

    def on_error(self, exception: Exception) -> None:
        """Handle an error from calling `read`."""


class Poller:
    """A poller to call a given reader on an interval.

    Args:
        reader: An interface to read data.
        interval: The poll interval, in seconds.
    """

    interval: float

    def __init__(self, reader: Reader, interval: float) -> None:
        self.interval = interval
        self._reader = reader
        self._read_lock: Optional["asyncio.Lock"] = None
        self._poll_waiters: List["asyncio.Future[None]"] = []
        self._poll_forever_task: Optional["asyncio.Task[None]"] = None

    async def start(self) -> None:
        if self._poll_forever_task is None:
            self._poll_forever_task = asyncio.create_task(self._poll_forever())
            await self.wait_next_poll()

    async def stop(self) -> None:
        """Stop polling."""
        task = self._poll_forever_task
        if task is not None:
            async with self._use_read_lock():
                task.cancel()
                await asyncio.gather(task, return_exceptions=True)
            for waiter in self._poll_waiters:
                waiter.cancel(msg="Module was removed")
            self._poll_forever_task = None

    async def wait_next_poll(self) -> None:
        """Wait for the next poll to complete.

        If called in the middle of a read, it will not return until
        the next complete read. If a read raises an exception,
        it will be passed through to `wait_next_poll`.
        """
        if not self._poll_forever_task or self._poll_forever_task.done():
            raise ModuleCommunicationError(message="Module was removed")

        poll_future = asyncio.get_running_loop().create_future()
        self._poll_waiters.append(poll_future)
        await poll_future

    @contextlib.asynccontextmanager
    async def _use_read_lock(self) -> AsyncGenerator[None, None]:
        self._read_lock = self._read_lock or asyncio.Lock()

        async with self._read_lock:
            yield

    async def _poll_forever(self) -> None:
        """Polling loop."""
        while True:
            await self._poll_once()
            await asyncio.sleep(self.interval)

    @staticmethod
    def _set_waiter_complete(
        waiter: "asyncio.Future[None]", e: Optional[Exception] = None
    ) -> None:
        try:
            waiter.set_result(None) if e is None else waiter.set_exception(e)
        except asyncio.InvalidStateError:
            log.warning("Poller waiter was already cancelled")

    async def _poll_once(self) -> None:
        """Trigger a single read, notifying listeners of success or error."""
        previous_waiters = self._poll_waiters
        self._poll_waiters = []

        try:
            async with self._use_read_lock():
                await self._reader.read()
        except asyncio.CancelledError:
            raise
        except AbsorbanceReaderDisconnectedError as e:
            for waiter in previous_waiters:
                Poller._set_waiter_complete(waiter, None)
            self._reader.on_error(e)
        except Exception as e:
            log.exception("Polling exception")
            self._reader.on_error(e)
            for waiter in previous_waiters:
                Poller._set_waiter_complete(waiter, e)
        else:
            for waiter in previous_waiters:
                Poller._set_waiter_complete(waiter)



"""Protocol engine types to do with liquid classes."""
from typing import Any
from pydantic import Field

from opentrons_shared_data.liquid_classes.liquid_class_definition import (
    ByTipTypeSetting,
)


class LiquidClassRecord(ByTipTypeSetting, frozen=True):
    """LiquidClassRecord is our internal representation of an (immutable) liquid class.

    Conceptually, a liquid class record is the tuple (name, pipette, tip, transfer properties).
    We consider two liquid classes to be the same if every entry in that tuple is the same; and liquid
    classes are different if any entry in the tuple is different.

    This class defines the tuple via inheritance so that we can reuse the definitions from shared_data.
    """

    liquidClassName: str = Field(
        ...,
        description="Identifier for the liquid of this liquid class, e.g. glycerol50.",
    )
    pipetteModel: str = Field(
        ...,
        description="Identifier for the pipette of this liquid class.",
    )
    # The other fields like tiprack ID, aspirate properties, etc. are pulled in from ByTipTypeSetting.

    def __hash__(self) -> int:
        """Hash function for LiquidClassRecord."""
        # Within the Protocol Engine, LiquidClassRecords are immutable, and we'd like to be able to
        # look up LiquidClassRecords by value, which involves hashing. However, Pydantic does not
        # generate a usable hash function if any of the subfields (like Coordinate) are not frozen.
        # So we have to implement the hash function ourselves.
        # Our strategy is to recursively convert this object into a list of (key, value) tuples.
        def dict_to_tuple(d: dict[str, Any]) -> tuple[tuple[str, Any], ...]:
            return tuple(
                (
                    field_name,
                    dict_to_tuple(value)
                    if isinstance(value, dict)
                    else tuple(value)
                    if isinstance(value, list)
                    else value,
                )
                for field_name, value in d.items()
            )

        return hash(dict_to_tuple(self.model_dump()))


class LiquidClassRecordWithId(LiquidClassRecord, frozen=True):
    """A LiquidClassRecord with its ID, for use in summary lists."""

    liquidClassId: str = Field(
        ...,
        description="Unique identifier for this liquid class.",
    )



from dataclasses import dataclass
from typing import List, Union, Sequence, Optional, Tuple

from opentrons.types import Location, NozzleMapInterface
from opentrons.protocols.api_support import instrument
from opentrons.protocols.advanced_control.transfers import (
    transfer_liquid_utils as tx_liquid_utils,
)
from opentrons.protocols.advanced_control.transfers.common import (
    TransferTipPolicyV2,
    TransferTipPolicyV2Type,
)

from .disposal_locations import TrashBin, WasteChute
from .labware import Labware, Well
from .core.common import WellCore
from . import validation


@dataclass
class TransferInfo:

    source: List[Well]
    dest: Union[List[Well], TrashBin, WasteChute]
    tip_policy: TransferTipPolicyV2
    tip_racks: List[Labware]
    trash_location: Union[Location, TrashBin, WasteChute]
    last_tip_location: Optional[Tuple[Location, WellCore]]


def verify_and_normalize_transfer_args(
    source: Union[Well, Sequence[Well], Sequence[Sequence[Well]]],
    dest: Union[Well, Sequence[Well], Sequence[Sequence[Well]], TrashBin, WasteChute],
    tip_policy: TransferTipPolicyV2Type,
    last_tip_well: Optional[Well],
    tip_racks: List[Labware],
    nozzle_map: NozzleMapInterface,
    group_wells_for_multi_channel: bool,
    current_volume: float,
    trash_location: Union[Location, Well, Labware, TrashBin, WasteChute],
) -> TransferInfo:
    flat_sources_list = validation.ensure_valid_flat_wells_list_for_transfer_v2(source)
    if not isinstance(dest, (TrashBin, WasteChute)):
        flat_dests_list = validation.ensure_valid_flat_wells_list_for_transfer_v2(dest)
    else:
        # If trash bin or waste chute, set this to empty to have less isinstance checks after this
        flat_dests_list = []
    if group_wells_for_multi_channel and nozzle_map.tip_count > 1:
        flat_sources_list = tx_liquid_utils.group_wells_for_multi_channel_transfer(
            flat_sources_list, nozzle_map, "source"
        )
        flat_dests_list = tx_liquid_utils.group_wells_for_multi_channel_transfer(
            flat_dests_list, nozzle_map, "destination"
        )
    for well in flat_sources_list + flat_dests_list:
        instrument.validate_takes_liquid(
            location=well.top(),
            reject_module=True,
            reject_adapter=True,
        )

    valid_new_tip = validation.ensure_new_tip_policy(tip_policy)
    if valid_new_tip == TransferTipPolicyV2.NEVER:
        if last_tip_well is None:
            raise RuntimeError(
                "Pipette has no tip attached to perform transfer."
                " Either do a pick_up_tip beforehand or specify a new_tip parameter"
                " of 'once' or 'always'."
            )
        else:
            valid_tip_racks = [last_tip_well.parent]
    else:
        valid_tip_racks = tip_racks
    if current_volume != 0:
        raise RuntimeError(
            "A transfer on a liquid class cannot start with liquid already in the tip."
            " Ensure that all previously aspirated liquid is dispensed before starting"
            " a new transfer."
        )

    _trash_location: Union[Location, Well, TrashBin, WasteChute]
    if isinstance(trash_location, Labware):
        _trash_location = trash_location.wells()[0]
    else:
        _trash_location = trash_location

    valid_trash_location = validation.ensure_valid_trash_location_for_transfer_v2(
        trash_location=_trash_location
    )

    if last_tip_well is not None:
        parent_tip_rack = last_tip_well.parent
        last_tip_location = (
            Location(last_tip_well.top().point, parent_tip_rack),
            last_tip_well._core,
        )
    else:
        last_tip_location = None

    return TransferInfo(
        source=flat_sources_list,
        dest=flat_dests_list if not isinstance(dest, (TrashBin, WasteChute)) else dest,
        tip_policy=valid_new_tip,
        tip_racks=valid_tip_racks,
        trash_location=valid_trash_location,
        last_tip_location=last_tip_location,
    )


def resolve_keep_last_tip(
    keep_last_tip: Optional[bool], tip_strategy: TransferTipPolicyV2
) -> bool:
    """Resolve the liquid class transfer argument `keep_last_tip`

    If set to a boolean value, maintains that setting. Otherwise, default to
    `True` if tip policy is `NEVER`, otherwise default to `False`
    """
    if keep_last_tip is not None:
        return keep_last_tip
    return tip_strategy == TransferTipPolicyV2.NEVER



"""Command models to initialize an Absorbance Reader."""
from __future__ import annotations
from typing import List, Optional, Literal, TYPE_CHECKING, Any
from typing_extensions import Type

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from opentrons.drivers.types import ABSMeasurementMode
from opentrons.protocol_engine.types import ABSMeasureMode

from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData
from ...errors.error_occurrence import ErrorOccurrence
from ...errors import InvalidWavelengthError
from ...state import update_types

if TYPE_CHECKING:
    from opentrons.protocol_engine.state.state import StateView
    from opentrons.protocol_engine.execution import EquipmentHandler


InitializeCommandType = Literal["absorbanceReader/initialize"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class InitializeParams(BaseModel):
    """Input parameters to initialize an absorbance reading."""

    moduleId: str = Field(..., description="Unique ID of the absorbance reader.")
    measureMode: ABSMeasureMode = Field(
        ..., description="Initialize single or multi measurement mode."
    )
    sampleWavelengths: List[int] = Field(..., description="Sample wavelengths in nm.")
    referenceWavelength: int | SkipJsonSchema[None] = Field(
        None,
        description="Optional reference wavelength in nm.",
        json_schema_extra=_remove_default,
    )


class InitializeResult(BaseModel):
    """Result data from initializing an aborbance reading."""


class InitializeImpl(
    AbstractCommandImpl[InitializeParams, SuccessData[InitializeResult]]
):
    """Execution implementation of initializing an Absorbance Reader."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment

    async def execute(self, params: InitializeParams) -> SuccessData[InitializeResult]:
        """Initiate a single absorbance measurement."""
        state_update = update_types.StateUpdate()
        abs_reader_substate = self._state_view.modules.get_absorbance_reader_substate(
            module_id=params.moduleId
        )
        # Allow propagation of ModuleNotAttachedError.
        abs_reader = self._equipment.get_module_hardware_api(
            abs_reader_substate.module_id
        )

        if abs_reader is not None:
            # Validate the parameters before initializing.
            sample_wavelengths = set(params.sampleWavelengths)
            sample_wavelengths_len = len(params.sampleWavelengths)
            reference_wavelength = params.referenceWavelength
            supported_wavelengths = set(abs_reader.supported_wavelengths)
            unsupported_wavelengths = sample_wavelengths.difference(
                supported_wavelengths
            )
            sample_wl_str = ", ".join([str(w) + "nm" for w in sample_wavelengths])
            supported_wl_str = ", ".join([str(w) + "nm" for w in supported_wavelengths])
            unsupported_wl_str = ", ".join(
                [str(w) + "nm" for w in unsupported_wavelengths]
            )
            if unsupported_wavelengths:
                raise InvalidWavelengthError(
                    f"Unsupported wavelengths: {unsupported_wl_str}. "
                    f" Use one of {supported_wl_str} instead."
                )

            if params.measureMode == "single":
                if sample_wavelengths_len != 1:
                    raise ValueError(
                        f"Measure mode `single` requires one sample wavelength,"
                        f" {sample_wl_str} provided instead."
                    )
                if (
                    reference_wavelength is not None
                    and reference_wavelength not in supported_wavelengths
                ):
                    raise InvalidWavelengthError(
                        f"Reference wavelength {reference_wavelength}nm is not supported."
                        f" Use one of {supported_wl_str} instead."
                    )

            if params.measureMode == "multi":
                if sample_wavelengths_len < 1 or sample_wavelengths_len > 6:
                    raise ValueError(
                        f"Measure mode `multi` requires 1-6 sample wavelengths,"
                        f" {sample_wl_str} provided instead."
                    )
                if reference_wavelength is not None:
                    raise ValueError(
                        "Reference wavelength cannot be used with Measure mode `multi`."
                    )

            await abs_reader.set_sample_wavelength(
                ABSMeasurementMode(params.measureMode),
                params.sampleWavelengths,
                reference_wavelength=params.referenceWavelength,
            )

        state_update.initialize_absorbance_reader(
            abs_reader_substate.module_id,
            params.measureMode,
            params.sampleWavelengths,
            params.referenceWavelength,
        )

        return SuccessData(public=InitializeResult(), state_update=state_update)


class Initialize(BaseCommand[InitializeParams, InitializeResult, ErrorOccurrence]):
    """A command to initialize an Absorbance Reader."""

    commandType: InitializeCommandType = "absorbanceReader/initialize"
    params: InitializeParams
    result: Optional[InitializeResult] = None

    _ImplementationCls: Type[InitializeImpl] = InitializeImpl


class InitializeCreate(BaseCommandCreate[InitializeParams]):
    """A request to execute an Absorbance Reader measurement."""

    commandType: InitializeCommandType = "absorbanceReader/initialize"
    params: InitializeParams

    _CommandCls: Type[Initialize] = Initialize



"""Protocol API module implementation logic."""

from __future__ import annotations

from typing import Optional, List, Dict, Union

from opentrons.hardware_control import SynchronousAdapter, modules as hw_modules
from opentrons.hardware_control.modules.types import (
    ModuleModel,
    TemperatureStatus,
    MagneticStatus,
    SpeedStatus,
    module_model_from_string,
)
from opentrons.drivers.types import (
    HeaterShakerLabwareLatchStatus,
    ThermocyclerLidStatus,
)

from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.types import ABSMeasureMode, StackerFillEmptyStrategy
from opentrons.types import DeckSlotName
from opentrons.protocol_engine.clients import SyncClient as ProtocolEngineClient
from opentrons.protocol_engine.errors.exceptions import (
    LabwareNotLoadedOnModuleError,
    NoMagnetEngageHeightError,
    CannotPerformModuleAction,
)

from opentrons.protocols.api_support.types import APIVersion, ThermocyclerStep

from ... import validation
from ..module import (
    AbstractModuleCore,
    AbstractTemperatureModuleCore,
    AbstractMagneticModuleCore,
    AbstractThermocyclerCore,
    AbstractHeaterShakerCore,
    AbstractMagneticBlockCore,
    AbstractAbsorbanceReaderCore,
    AbstractFlexStackerCore,
)
from .exceptions import InvalidMagnetEngageHeightError
from . import load_labware_params

# Valid wavelength range for absorbance reader
ABS_WAVELENGTH_MIN = 350
ABS_WAVELENGTH_MAX = 1000


class ModuleCore(AbstractModuleCore):
    """Module core logic implementation for Python protocols.
    Args:
        module_id: ProtocolEngine ID of the loaded modules.
    """

    def __init__(
        self,
        module_id: str,
        engine_client: ProtocolEngineClient,
        api_version: APIVersion,
        sync_module_hardware: SynchronousAdapter[hw_modules.AbstractModule],
    ) -> None:
        self._module_id = module_id
        self._engine_client = engine_client
        self._api_version = api_version
        self._sync_module_hardware = sync_module_hardware

    @property
    def api_version(self) -> APIVersion:
        """Get the api version protocol module target."""
        return self._api_version

    @property
    def module_id(self) -> str:
        """The module's unique ProtocolEngine ID."""
        return self._module_id

    def get_model(self) -> ModuleModel:
        """Get the module's model identifier."""
        return module_model_from_string(
            self._engine_client.state.modules.get_connected_model(self.module_id)
        )

    def get_serial_number(self) -> str:
        """Get the module's unique hardware serial number."""
        return self._engine_client.state.modules.get_serial_number(self.module_id)

    def get_deck_slot(self) -> DeckSlotName:
        """Get the module's deck slot."""
        return self._engine_client.state.modules.get_location(self.module_id).slotName

    def get_deck_slot_id(self) -> str:
        slot_name = self.get_deck_slot()
        return validation.internal_slot_to_public_string(
            slot_name, robot_type=self._engine_client.state.config.robot_type
        )

    def get_display_name(self) -> str:
        """Get the module's display name."""
        return self._engine_client.state.modules.get_definition(
            self.module_id
        ).displayName


class NonConnectedModuleCore(AbstractModuleCore):
    """Not connected module core logic implementation for Python protocols.

    Args:
        module_id: ProtocolEngine ID of the loaded modules.
    """

    def __init__(
        self,
        module_id: str,
        engine_client: ProtocolEngineClient,
        api_version: APIVersion,
    ) -> None:
        self._module_id = module_id
        self._engine_client = engine_client
        self._api_version = api_version

    @property
    def api_version(self) -> APIVersion:
        """Get the api version protocol module target."""
        return self._api_version

    @property
    def module_id(self) -> str:
        """The module's unique ProtocolEngine ID."""
        return self._module_id

    def get_model(self) -> ModuleModel:
        """Get the module's model identifier."""
        return module_model_from_string(
            self._engine_client.state.modules.get_connected_model(self.module_id)
        )

    def get_deck_slot(self) -> DeckSlotName:
        """Get the module's deck slot."""
        return self._engine_client.state.modules.get_location(self.module_id).slotName

    def get_display_name(self) -> str:
        """Get the module's display name."""
        return self._engine_client.state.modules.get_definition(
            self.module_id
        ).displayName

    def get_deck_slot_id(self) -> str:
        slot_name = self.get_deck_slot()
        return validation.internal_slot_to_public_string(
            slot_name, robot_type=self._engine_client.state.config.robot_type
        )


class TemperatureModuleCore(ModuleCore, AbstractTemperatureModuleCore):
    """Temperature Module core logic implementation for Python protocols."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.TempDeck]

    def set_target_temperature(self, celsius: float) -> None:
        """Set the Temperature Module's target temperature in °C."""
        self._engine_client.execute_command(
            cmd.temperature_module.SetTargetTemperatureParams(
                moduleId=self.module_id, celsius=celsius
            )
        )

    def wait_for_target_temperature(self, celsius: Optional[float] = None) -> None:
        """Wait until the module's target temperature is reached.
        Specifying a value for ``celsius`` that is different than
        the module's current target temperature may behave unpredictably.
        """
        self._engine_client.execute_command(
            cmd.temperature_module.WaitForTemperatureParams(
                moduleId=self.module_id, celsius=celsius
            )
        )

    def deactivate(self) -> None:
        """Deactivate the Temperature Module."""
        self._engine_client.execute_command(
            cmd.temperature_module.DeactivateTemperatureParams(moduleId=self.module_id)
        )

    def get_current_temperature(self) -> float:
        """Get the module's current temperature in °C."""
        return self._sync_module_hardware.temperature  # type: ignore[no-any-return]

    def get_target_temperature(self) -> Optional[float]:
        """Get the module's target temperature in °C, if set."""
        return self._sync_module_hardware.target  # type: ignore[no-any-return]

    def get_status(self) -> TemperatureStatus:
        """Get the module's current temperature status."""
        return self._sync_module_hardware.status  # type: ignore[no-any-return]


class MagneticModuleCore(ModuleCore, AbstractMagneticModuleCore):
    """Magnetic Module control interface via a ProtocolEngine."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.MagDeck]

    def engage(
        self,
        height_from_base: Optional[float] = None,
        height_from_home: Optional[float] = None,
    ) -> None:
        """Raise the module's magnets.
        Only one of `height_from_base` or `height_from_home` may be specified.
        Args:
            height_from_base: Distance from labware base to raise the magnets.
            height_from_home: Distance from motor home position to raise the magnets.
        """

        # This core will only be used in apiLevels >=2.14, where
        # MagneticModuleContext.engage(height=...) is no longer available.
        # So these asserts should always pass.
        assert (
            height_from_home is None
        ), "Expected engage height to be specified from base."
        assert (
            height_from_base is not None
        ), "Expected engage height to be specified from base."

        self._engine_client.execute_command(
            cmd.magnetic_module.EngageParams(
                moduleId=self._module_id, height=height_from_base
            )
        )

    def engage_to_labware(
        self, offset: float = 0, preserve_half_mm: bool = False
    ) -> None:
        """Raise the module's magnets up to its loaded labware.
        Args:
            offset: Offset from the labware's default engage height.
            preserve_half_mm: For labware whose definitions
                erroneously use half-mm for their defined default engage height,
                use the value directly instead of converting it to real millimeters.
        """
        try:
            default_height = (
                self._engine_client.state.labware.get_default_magnet_height(
                    module_id=self.module_id, offset=offset
                )
            )
        except LabwareNotLoadedOnModuleError:
            raise InvalidMagnetEngageHeightError(
                "There is no labware loaded on this Magnetic Module,"
                " so you must specify an engage height"
                " with the `height_from_base` parameter."
            )
        except NoMagnetEngageHeightError:
            raise InvalidMagnetEngageHeightError(
                "The labware loaded on this Magnetic Module"
                " does not have a default engage height,"
                " so you must specify an engage height"
                " with the `height_from_base` parameter."
            )

        self._engine_client.execute_command(
            cmd.magnetic_module.EngageParams(
                moduleId=self.module_id, height=default_height
            )
        )

    def disengage(self) -> None:
        """Lower the magnets back into the module."""
        self._engine_client.execute_command(
            cmd.magnetic_module.DisengageParams(moduleId=self.module_id)
        )

    def get_status(self) -> MagneticStatus:
        """Get the module's current magnet status."""
        return self._sync_module_hardware.status  # type: ignore[no-any-return]


class ThermocyclerModuleCore(ModuleCore, AbstractThermocyclerCore):
    """Core control interface for an attached Thermocycler Module."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.Thermocycler]
    _repetitions: Optional[int] = None
    _step_count: Optional[int] = None

    def open_lid(self) -> ThermocyclerLidStatus:
        """Open the Thermocycler's lid."""
        self._engine_client.execute_command(
            cmd.thermocycler.OpenLidParams(moduleId=self.module_id)
        )
        return ThermocyclerLidStatus.OPEN

    def close_lid(self) -> ThermocyclerLidStatus:
        """Close the Thermocycler's lid."""
        self._engine_client.execute_command(
            cmd.thermocycler.CloseLidParams(moduleId=self.module_id)
        )
        return ThermocyclerLidStatus.CLOSED

    def set_target_block_temperature(
        self,
        celsius: float,
        hold_time_seconds: Optional[float] = None,
        block_max_volume: Optional[float] = None,
    ) -> None:
        """Set the target temperature for the well block, in °C."""
        self._engine_client.execute_command(
            cmd.thermocycler.SetTargetBlockTemperatureParams(
                moduleId=self.module_id,
                celsius=celsius,
                blockMaxVolumeUl=block_max_volume,
                holdTimeSeconds=hold_time_seconds,
            )
        )

    def wait_for_block_temperature(self) -> None:
        """Wait for target block temperature to be reached."""
        self._engine_client.execute_command(
            cmd.thermocycler.WaitForBlockTemperatureParams(moduleId=self.module_id)
        )

    def set_target_lid_temperature(self, celsius: float) -> None:
        """Set the target temperature for the heated lid, in °C."""
        self._engine_client.execute_command(
            cmd.thermocycler.SetTargetLidTemperatureParams(
                moduleId=self.module_id, celsius=celsius
            )
        )

    def wait_for_lid_temperature(self) -> None:
        """Wait for target lid temperature to be reached."""
        self._engine_client.execute_command(
            cmd.thermocycler.WaitForLidTemperatureParams(moduleId=self.module_id)
        )

    def _execute_profile_pre_221(
        self,
        steps: List[ThermocyclerStep],
        repetitions: int,
        block_max_volume: Optional[float],
    ) -> None:
        """Execute a thermocycler profile using thermocycler/runProfile and flattened steps."""
        engine_steps = [
            cmd.thermocycler.RunProfileStepParams(
                celsius=step["temperature"],
                holdSeconds=step["hold_time_seconds"],
            )
            for step in steps
        ]
        repeated_engine_steps = engine_steps * repetitions
        self._engine_client.execute_command(
            cmd.thermocycler.RunProfileParams(
                moduleId=self.module_id,
                profile=repeated_engine_steps,
                blockMaxVolumeUl=block_max_volume,
            )
        )

    def _execute_profile_post_221(
        self,
        steps: List[ThermocyclerStep],
        repetitions: int,
        block_max_volume: Optional[float],
    ) -> None:
        """Execute a thermocycler profile using thermocycler/runExtendedProfile."""
        engine_steps: List[
            Union[cmd.thermocycler.ProfileCycle, cmd.thermocycler.ProfileStep]
        ] = [
            cmd.thermocycler.ProfileCycle(
                repetitions=repetitions,
                steps=[
                    cmd.thermocycler.ProfileStep(
                        celsius=step["temperature"],
                        holdSeconds=step["hold_time_seconds"],
                    )
                    for step in steps
                ],
            )
        ]
        self._engine_client.execute_command(
            cmd.thermocycler.RunExtendedProfileParams(
                moduleId=self.module_id,
                profileElements=engine_steps,
                blockMaxVolumeUl=block_max_volume,
            )
        )

    def execute_profile(
        self,
        steps: List[ThermocyclerStep],
        repetitions: int,
        block_max_volume: Optional[float] = None,
    ) -> None:
        """Execute a Thermocycler Profile."""
        self._repetitions = repetitions
        self._step_count = len(steps)
        if self.api_version >= APIVersion(2, 21):
            return self._execute_profile_post_221(steps, repetitions, block_max_volume)
        else:
            return self._execute_profile_pre_221(steps, repetitions, block_max_volume)

    def deactivate_lid(self) -> None:
        """Turn off the heated lid."""
        self._engine_client.execute_command(
            cmd.thermocycler.DeactivateLidParams(moduleId=self.module_id)
        )

    def deactivate_block(self) -> None:
        """Turn off the well block temperature controller"""
        self._clear_cycle_counters()
        self._engine_client.execute_command(
            cmd.thermocycler.DeactivateBlockParams(moduleId=self.module_id)
        )

    def deactivate(self) -> None:
        """Turn off the well block temperature controller, and heated lid"""
        self.deactivate_block()
        self.deactivate_lid()

    def get_lid_position(self) -> Optional[ThermocyclerLidStatus]:
        """Get the thermocycler's lid position."""
        return self._sync_module_hardware.lid_status  # type: ignore[no-any-return]

    def get_block_temperature_status(self) -> TemperatureStatus:
        """Get the thermocycler's block temperature status."""
        return self._sync_module_hardware.status  # type: ignore[no-any-return]

    def get_lid_temperature_status(self) -> Optional[TemperatureStatus]:
        """Get the thermocycler's lid temperature status."""
        return self._sync_module_hardware.lid_temp_status  # type: ignore[no-any-return]

    def get_block_temperature(self) -> Optional[float]:
        """Get the thermocycler's current block temperature in °C."""
        return self._sync_module_hardware.temperature  # type: ignore[no-any-return]

    def get_block_target_temperature(self) -> Optional[float]:
        """Get the thermocycler's target block temperature in °C."""
        return self._sync_module_hardware.target  # type: ignore[no-any-return]

    def get_lid_temperature(self) -> Optional[float]:
        """Get the thermocycler's current lid temperature in °C."""
        return self._sync_module_hardware.lid_temp  # type: ignore[no-any-return]

    def get_lid_target_temperature(self) -> Optional[float]:
        """Get the thermocycler's target lid temperature in °C."""
        return self._sync_module_hardware.lid_target  # type: ignore[no-any-return]

    def get_ramp_rate(self) -> Optional[float]:
        """Get the thermocycler's current ramp rate in °C/sec."""
        return self._sync_module_hardware.ramp_rate  # type: ignore[no-any-return]

    def get_hold_time(self) -> Optional[float]:
        """Get the remaining hold time in seconds."""
        return self._sync_module_hardware.hold_time  # type: ignore[no-any-return]

    def get_total_cycle_count(self) -> Optional[int]:
        """Get number of repetitions for current set cycle."""
        return self._repetitions

    def get_current_cycle_index(self) -> Optional[int]:
        """Get index of the current set cycle repetition."""
        if self._repetitions is None:
            return None
        step_index = self._sync_module_hardware.current_step_index
        # TODO(jbl 2022-10-31) this is intended to work even if execute profile is non-blocking, but it is blocking so
        #   this is not guaranteed to be accurate
        return (step_index - 1) // self._step_count + 1  # type: ignore[no-any-return]

    def get_total_step_count(self) -> Optional[int]:
        """Get number of steps within the current cycle."""
        return self._step_count

    def get_current_step_index(self) -> Optional[int]:
        """Get the index of the current step within the current cycle."""
        if self._step_count is None:
            return None
        step_index = self._sync_module_hardware.current_step_index
        # TODO(jbl 2022-10-31) this is intended to work even if execute profile is non-blocking, but it is blocking so
        #   this is not guaranteed to be accurate
        return (step_index - 1) % self._step_count + 1  # type: ignore[no-any-return]

    def _clear_cycle_counters(self) -> None:
        """Clear core-tracked cycle counters."""
        self._repetitions = None
        self._step_count = None


class HeaterShakerModuleCore(ModuleCore, AbstractHeaterShakerCore):
    """Core control interface for an attached Heater-Shaker Module."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.HeaterShaker]

    def set_target_temperature(self, celsius: float) -> None:
        """Set the labware plate's target temperature in °C."""
        self._engine_client.execute_command(
            cmd.heater_shaker.SetTargetTemperatureParams(
                moduleId=self.module_id, celsius=celsius
            )
        )

    def wait_for_target_temperature(self) -> None:
        """Wait for the labware plate's target temperature to be reached."""
        self._engine_client.execute_command(
            cmd.heater_shaker.WaitForTemperatureParams(moduleId=self.module_id)
        )

    def set_and_wait_for_shake_speed(self, rpm: int) -> None:
        """Set the shaker's target shake speed and wait for it to spin up."""
        self._engine_client.execute_command(
            cmd.heater_shaker.SetAndWaitForShakeSpeedParams(
                moduleId=self.module_id, rpm=rpm
            )
        )

    def open_labware_latch(self) -> None:
        """Open the labware latch."""
        self._engine_client.execute_command(
            cmd.heater_shaker.OpenLabwareLatchParams(moduleId=self.module_id)
        )

    def close_labware_latch(self) -> None:
        """Close the labware latch."""
        self._engine_client.execute_command(
            cmd.heater_shaker.CloseLabwareLatchParams(moduleId=self.module_id)
        )

    def deactivate_shaker(self) -> None:
        """Stop shaking."""
        self._engine_client.execute_command(
            cmd.heater_shaker.DeactivateShakerParams(moduleId=self.module_id)
        )

    def deactivate_heater(self) -> None:
        """Stop heating."""
        self._engine_client.execute_command(
            cmd.heater_shaker.DeactivateHeaterParams(moduleId=self.module_id)
        )

    def get_current_temperature(self) -> float:
        """Get the labware plate's current temperature in °C."""
        return self._sync_module_hardware.temperature  # type: ignore[no-any-return]

    def get_target_temperature(self) -> Optional[float]:
        """Get the labware plate's target temperature in °C, if set."""
        return self._sync_module_hardware.target_temperature  # type: ignore[no-any-return]

    def get_current_speed(self) -> int:
        """Get the shaker's current speed in RPM."""
        return self._sync_module_hardware.speed  # type: ignore[no-any-return]

    def get_target_speed(self) -> Optional[int]:
        """Get the shaker's target speed in RPM, if set."""
        return self._sync_module_hardware.target_speed  # type: ignore[no-any-return]

    def get_temperature_status(self) -> TemperatureStatus:
        """Get the module's heater status."""
        return self._sync_module_hardware.temperature_status  # type: ignore[no-any-return]

    def get_speed_status(self) -> SpeedStatus:
        """Get the module's heater status."""
        return self._sync_module_hardware.speed_status  # type: ignore[no-any-return]

    def get_labware_latch_status(self) -> HeaterShakerLabwareLatchStatus:
        """Get the module's labware latch status."""
        return self._sync_module_hardware.labware_latch_status  # type: ignore[no-any-return]


class MagneticBlockCore(NonConnectedModuleCore, AbstractMagneticBlockCore):
    """Magnetic Block control interface via a ProtocolEngine."""


class AbsorbanceReaderCore(ModuleCore, AbstractAbsorbanceReaderCore):
    """Absorbance Reader core logic implementation for Python protocols."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.AbsorbanceReader]
    _initialized_value: Optional[List[int]] = None
    _ready_to_initialize: bool = False

    def initialize(
        self,
        mode: ABSMeasureMode,
        wavelengths: List[int],
        reference_wavelength: Optional[int] = None,
    ) -> None:
        """Initialize the Absorbance Reader by taking zero reading."""
        if not self._ready_to_initialize:
            raise CannotPerformModuleAction(
                "Cannot perform Initialize action on Absorbance Reader without calling `.close_lid()` first."
            )

        wavelength_len = len(wavelengths)
        if mode == "single" and wavelength_len != 1:
            raise ValueError(
                f"Single mode can only be initialized with 1 wavelength"
                f" {wavelength_len} wavelengths provided instead."
            )

        if mode == "multi" and (wavelength_len < 1 or wavelength_len > 6):
            raise ValueError(
                f"Multi mode can only be initialized with 1 - 6 wavelengths."
                f" {wavelength_len} wavelengths provided instead."
            )

        if reference_wavelength is not None and (
            reference_wavelength < ABS_WAVELENGTH_MIN
            or reference_wavelength > ABS_WAVELENGTH_MAX
        ):
            raise ValueError(
                f"Unsupported reference wavelength: ({reference_wavelength}) needs"
                f" to between {ABS_WAVELENGTH_MIN} and {ABS_WAVELENGTH_MAX} nm."
            )

        for wavelength in wavelengths:
            if (
                not isinstance(wavelength, int)
                or wavelength < ABS_WAVELENGTH_MIN
                or wavelength > ABS_WAVELENGTH_MAX
            ):
                raise ValueError(
                    f"Unsupported sample wavelength: ({wavelength}) needs"
                    f" to between {ABS_WAVELENGTH_MIN} and {ABS_WAVELENGTH_MAX} nm."
                )

        self._engine_client.execute_command(
            cmd.absorbance_reader.InitializeParams(
                moduleId=self.module_id,
                measureMode=mode,
                sampleWavelengths=wavelengths,
                referenceWavelength=reference_wavelength,
            ),
        )
        self._initialized_value = wavelengths

    def read(self, filename: Optional[str] = None) -> Dict[int, Dict[str, float]]:
        """Initiate a read on the Absorbance Reader, and return the results. During Analysis, this will return a measurement of zero for all wells."""
        wavelengths = self._engine_client.state.modules.get_absorbance_reader_substate(
            self.module_id
        ).configured_wavelengths
        if wavelengths is None:
            raise CannotPerformModuleAction(
                "Cannot perform Read action on Absorbance Reader without calling `.initialize(...)` first."
            )
        if self._initialized_value:
            self._engine_client.execute_command(
                cmd.absorbance_reader.ReadAbsorbanceParams(
                    moduleId=self.module_id, fileName=filename
                )
            )
        if not self._engine_client.state.config.use_virtual_modules:
            read_result = (
                self._engine_client.state.modules.get_absorbance_reader_substate(
                    self.module_id
                ).data
            )
            if read_result is not None:
                return read_result
            raise CannotPerformModuleAction(
                "Absorbance Reader failed to return expected read result."
            )

        # When using virtual modules, return all zeroes
        virtual_asbsorbance_result: Dict[int, Dict[str, float]] = {}
        for wavelength in wavelengths:
            converted_values = (
                self._engine_client.state.modules.convert_absorbance_reader_data_points(
                    data=[0] * 96
                )
            )
            virtual_asbsorbance_result[wavelength] = converted_values
        return virtual_asbsorbance_result

    def close_lid(
        self,
    ) -> None:
        """Close the Absorbance Reader's lid."""
        self._engine_client.execute_command(
            cmd.absorbance_reader.CloseLidParams(
                moduleId=self.module_id,
            )
        )
        self._ready_to_initialize = True

    def open_lid(self) -> None:
        """Close the Absorbance Reader's lid."""
        self._engine_client.execute_command(
            cmd.absorbance_reader.OpenLidParams(
                moduleId=self.module_id,
            )
        )

    def is_lid_on(self) -> bool:
        """Returns True if the Absorbance Reader's lid is currently on the Reader slot."""
        abs_state = self._engine_client.state.modules.get_absorbance_reader_substate(
            self.module_id
        )
        return abs_state.is_lid_on


class FlexStackerCore(ModuleCore, AbstractFlexStackerCore):
    """Flex Stacker core logic implementation for Python protocols."""

    _sync_module_hardware: SynchronousAdapter[hw_modules.FlexStacker]

    def retrieve(self) -> None:
        """Retrieve a labware from the Flex Stacker's hopper."""
        self._engine_client.execute_command(
            cmd.flex_stacker.RetrieveParams(
                moduleId=self.module_id,
            )
        )

    def store(self) -> None:
        """Store a labware into Flex Stacker's hopper."""
        self._engine_client.execute_command(
            cmd.flex_stacker.StoreParams(
                moduleId=self.module_id,
            )
        )

    def fill(self, message: str, count: int | None) -> None:
        """Pause the protocol to add more labware to the Flex Stacker's hopper."""
        self._engine_client.execute_command(
            cmd.flex_stacker.FillParams(
                moduleId=self.module_id,
                strategy=StackerFillEmptyStrategy.MANUAL_WITH_PAUSE,
                message=message,
                count=count,
            )
        )

    def empty(self, message: str) -> None:
        """Pause the protocol to remove labware from the Flex Stacker's hopper."""
        self._engine_client.execute_command(
            cmd.flex_stacker.EmptyParams(
                moduleId=self.module_id,
                strategy=StackerFillEmptyStrategy.MANUAL_WITH_PAUSE,
                message=message,
                count=0,
            )
        )

    def set_stored_labware(
        self,
        main_load_name: str,
        main_namespace: str | None,
        main_version: int | None,
        lid_load_name: str | None,
        lid_namespace: str | None,
        lid_version: int | None,
        adapter_load_name: str | None,
        adapter_namespace: str | None,
        adapter_version: int | None,
        count: int | None,
    ) -> None:
        """Configure the kind of labware that the stacker stores."""

        custom_labware_params = (
            self._engine_client.state.labware.find_custom_labware_load_params()
        )

        main_namespace, main_version = load_labware_params.resolve(
            main_load_name,
            main_namespace,
            main_version,
            custom_labware_params,
            self._api_version,
        )
        main_labware = cmd.flex_stacker.StackerStoredLabwareDetails(
            loadName=main_load_name, namespace=main_namespace, version=main_version
        )

        lid_labware: cmd.flex_stacker.StackerStoredLabwareDetails | None = None

        if lid_load_name:
            lid_namespace, lid_version = load_labware_params.resolve(
                lid_load_name,
                lid_namespace,
                lid_version,
                custom_labware_params,
                self._api_version,
            )
            lid_labware = cmd.flex_stacker.StackerStoredLabwareDetails(
                loadName=lid_load_name, namespace=lid_namespace, version=lid_version
            )

        adapter_labware: cmd.flex_stacker.StackerStoredLabwareDetails | None = None

        if adapter_load_name:
            adapter_namespace, adapter_version = load_labware_params.resolve(
                adapter_load_name,
                adapter_namespace,
                adapter_version,
                custom_labware_params,
                self._api_version,
            )
            adapter_labware = cmd.flex_stacker.StackerStoredLabwareDetails(
                loadName=adapter_load_name,
                namespace=adapter_namespace,
                version=adapter_version,
            )

        self._engine_client.execute_command(
            cmd.flex_stacker.SetStoredLabwareParams(
                moduleId=self.module_id,
                initialCount=count,
                primaryLabware=main_labware,
                lidLabware=lid_labware,
                adapterLabware=adapter_labware,
            )
        )



from __future__ import annotations

import asyncio
import logging
from typing import Callable, Optional, List, Dict, Mapping, Union, cast
from opentrons.drivers.rpi_drivers.types import USBPort
from opentrons.drivers.types import ThermocyclerLidStatus, Temperature, PlateTemperature
from opentrons.hardware_control.modules.lid_temp_status import LidTemperatureStatus
from opentrons.hardware_control.modules.plate_temp_status import PlateTemperatureStatus
from opentrons.hardware_control.modules.types import (
    ModuleDisconnectedCallback,
    TemperatureStatus,
)
from opentrons.hardware_control.poller import Reader, Poller

from ..execution_manager import ExecutionManager
from . import types, update, mod_abc
from opentrons.drivers.thermocycler import (
    AbstractThermocyclerDriver,
    SimulatingDriver,
    ThermocyclerDriverV2,
    ThermocyclerDriverFactory,
)


log = logging.getLogger(__name__)

POLLING_FREQUENCY_SEC = 1.0
SIM_POLLING_FREQUENCY_SEC = POLLING_FREQUENCY_SEC / 50.0

V1_MODULE_STRING = "thermocyclerModuleV1"
V2_MODULE_STRING = "thermocyclerModuleV2"

DFU_PID = "df11"

_TC_PLATE_LIFT_OPEN_DEGREES = 20
_TC_PLATE_LIFT_RETURN_DEGREES = 23


class ThermocyclerError(Exception):
    pass


def _temperature_is_holding(status: Optional[TemperatureStatus]) -> bool:
    if status in (TemperatureStatus.HOLDING, TemperatureStatus.IDLE):
        return True

    if status == TemperatureStatus.ERROR:
        raise ThermocyclerError("Error occurred while waiting for temperature")

    return False


class Thermocycler(mod_abc.AbstractModule):
    """Hardware control interface for an attached Thermocycler."""

    MODULE_TYPE = types.ModuleType.THERMOCYCLER

    @classmethod
    async def build(
        cls,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        hw_control_loop: asyncio.AbstractEventLoop,
        poll_interval_seconds: Optional[float] = None,
        simulating: bool = False,
        sim_model: Optional[str] = None,
        sim_serial_number: Optional[str] = None,
        disconnected_callback: ModuleDisconnectedCallback = None,
    ) -> "Thermocycler":
        """
        Build and connect to a Thermocycler

        Args:
            port: The port to connect to
            usb_port: USB Port
            execution_manager: Execution manager.
            hw_control_loop: The event loop running in the hardware control thread.
            poll_interval_seconds: Poll interval override.
            simulating: whether to build a simulating driver
            loop: Loop
            sim_model: The model name used by simulator
            disconnected_callback: Callback to inform the module controller that the device was disconnected

        Returns:
            Thermocycler instance.
        """
        driver: AbstractThermocyclerDriver
        if not simulating:
            driver = await ThermocyclerDriverFactory.create(
                port=port, loop=hw_control_loop
            )
            poll_interval_seconds = poll_interval_seconds or POLLING_FREQUENCY_SEC
        else:
            driver = SimulatingDriver(model=sim_model, serial_number=sim_serial_number)
            poll_interval_seconds = poll_interval_seconds or SIM_POLLING_FREQUENCY_SEC

        reader = ThermocyclerReader(driver=driver)
        poller = Poller(reader=reader, interval=poll_interval_seconds)
        module = cls(
            port=port,
            usb_port=usb_port,
            driver=driver,
            reader=reader,
            poller=poller,
            device_info=await driver.get_device_info(),
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
        )

        try:
            await poller.start()
        except Exception:
            log.exception(f"First read of Thermocycler on port {port} failed")

        return module

    def __init__(
        self,
        port: str,
        usb_port: USBPort,
        execution_manager: ExecutionManager,
        driver: AbstractThermocyclerDriver,
        reader: ThermocyclerReader,
        poller: Poller,
        device_info: Dict[str, str],
        hw_control_loop: asyncio.AbstractEventLoop,
        disconnected_callback: ModuleDisconnectedCallback = None,
    ) -> None:
        """
        Constructor

        Args:
            port: The port the thermocycler is connected to.
            usb_port: The USB port.
            execution_manager: The hardware execution manager.
            driver: The thermocycler driver.
            reader: An interface to read data from the Thermocycler.
            poller: A poll controller for reads.
            device_info: The thermocycler device info.
            hw_control_loop: The event loop running in the hardware control thread.
            disconnected_callback: Callback to inform the module controller that the device was disconnected
        """
        self._driver = driver
        super().__init__(
            port=port,
            usb_port=usb_port,
            hw_control_loop=hw_control_loop,
            execution_manager=execution_manager,
            disconnected_callback=disconnected_callback,
        )
        self._device_info = device_info
        self._reader = reader
        self._poller = poller
        self._total_cycle_count: Optional[int] = None
        self._current_cycle_index: Optional[int] = None
        self._total_step_count: Optional[int] = None
        self._current_step_index: Optional[int] = None
        self._error: Optional[str] = None
        self._reader.register_error_handler(self._enter_error_state)

    async def cleanup(self) -> None:
        """Stop the poller task."""
        await self._poller.stop()
        await self._driver.disconnect()

    @classmethod
    def name(cls) -> str:
        return "thermocycler"

    def firmware_prefix(self) -> str:
        """The prefix used for looking up firmware"""
        if self.model() == V1_MODULE_STRING:
            return "thermocycler"
        else:
            return "thermocycler-gen2"

    def model(self) -> str:
        if isinstance(self._driver, SimulatingDriver):
            return self._driver.model()
        elif isinstance(self._driver, ThermocyclerDriverV2):
            return V2_MODULE_STRING
        else:
            # Real module that is not a V2
            return V1_MODULE_STRING

    def bootloader(self) -> types.UploadFunction:
        if self.model() == V2_MODULE_STRING:
            return update.upload_via_dfu
        else:
            return update.upload_via_bossa

    def _clear_cycle_counters(self) -> None:
        """Clear the cycle counters."""
        self._total_cycle_count = None
        self._current_cycle_index = None
        self._total_step_count = None
        self._current_step_index = None

    async def deactivate_lid(self, must_be_running: bool = True) -> None:
        """Deactivate the lid heating pad"""
        if must_be_running:
            await self.wait_for_is_running()
        await self._driver.deactivate_lid()
        await self._reader.read_lid_temperature()

    async def deactivate_block(self, must_be_running: bool = True) -> None:
        """Deactivate the block peltiers"""
        if must_be_running:
            await self.wait_for_is_running()
        self._clear_cycle_counters()
        await self._driver.deactivate_block()
        await self._reader.read_block_temperature()

    async def deactivate(self, must_be_running: bool = True) -> None:
        """Deactivate the block peltiers and lid heating pad"""
        if must_be_running:
            await self.wait_for_is_running()
        self._clear_cycle_counters()
        await self._driver.deactivate_all()
        await self._reader.read()

    async def open(self) -> str:
        """Open the lid if it is closed"""
        await self.wait_for_is_running()
        await self._driver.open_lid()
        await self._wait_for_lid_status(ThermocyclerLidStatus.OPEN)
        return ThermocyclerLidStatus.OPEN

    async def close(self) -> str:
        """Close the lid if it is open"""
        await self.wait_for_is_running()
        await self._driver.close_lid()
        await self._wait_for_lid_status(ThermocyclerLidStatus.CLOSED)
        return ThermocyclerLidStatus.CLOSED

    async def lift_plate(self) -> str:
        """If the lid is open, lift the plate.

        Note that this should always be preceded by an `open` call. If the
        lid is not open, the Thermocycler will respond with an error message.

        Returns: Thermocycler lid status after running the command.
        """
        await self.wait_for_is_running()
        await self._driver.lift_plate()
        await self._wait_for_lid_status(ThermocyclerLidStatus.OPEN)
        return ThermocyclerLidStatus.OPEN

    async def raise_plate(self) -> None:
        """Move lid an extra bit."""
        if not self.lid_status == ThermocyclerLidStatus.OPEN:
            raise RuntimeError("Lid must be open")
        await self.wait_for_is_running()
        await self._driver.jog_lid(_TC_PLATE_LIFT_OPEN_DEGREES)
        # There is NO WAIT for the updated lid stats, because it will always
        # remain at "open"

    async def return_from_raise_plate(self) -> None:
        """Return lid back to normal open position."""
        await self.wait_for_is_running()
        await self._driver.jog_lid(-_TC_PLATE_LIFT_RETURN_DEGREES)
        await self.open()
        await self._wait_for_lid_status(ThermocyclerLidStatus.OPEN)

    async def set_temperature(
        self,
        temperature: float,
        hold_time_seconds: Optional[float] = None,
        hold_time_minutes: Optional[float] = None,
        ramp_rate: Optional[float] = None,
        volume: Optional[float] = None,
    ) -> None:
        """
        Set the temperature and wait.

        If hold time is set this function will return after
        the hold time expires.

        Otherwise it will return when the target temperature is reached.

        Args:
            temperature: The target temperature.
            hold_time_seconds: Optional number of seconds to wait.
            hold_time_minutes: Optional number of minutes to wait.
            ramp_rate: Optional ramp rate.
            volume: Optional volume.

        Returns: None
        """
        await self.wait_for_is_running()
        await self._set_temperature_no_pause(
            temperature=temperature,
            hold_time_seconds=hold_time_seconds,
            hold_time_minutes=hold_time_minutes,
            ramp_rate=ramp_rate,
            volume=volume,
        )

    async def _set_temperature_no_pause(
        self,
        temperature: float,
        hold_time_seconds: Optional[float],
        hold_time_minutes: Optional[float],
        ramp_rate: Optional[float],
        volume: Optional[float],
    ) -> None:
        seconds = hold_time_seconds if hold_time_seconds is not None else 0
        minutes = hold_time_minutes if hold_time_minutes is not None else 0
        total_seconds = seconds + (minutes * 60)
        hold_time = total_seconds if total_seconds > 0 else 0

        if ramp_rate is not None:
            await self._driver.set_ramp_rate(ramp_rate=ramp_rate)

        await self._driver.set_plate_temperature(
            temp=temperature, hold_time=hold_time, volume=volume
        )

        task = self._loop.create_task(self._wait_for_block_target())
        self.make_cancellable(task)
        await task

    async def wait_for_block_target(self) -> None:
        """
        Wait for thermocycler to reach given temperature.

        Will return when the target temperature is reached.

        Args:
            temperature: The target temperature.

        Returns: None
        """
        await self.wait_for_is_running()
        task = self._loop.create_task(self._wait_for_block_target())
        self.make_cancellable(task)
        await task

    async def cycle_temperatures(
        self,
        steps: List[types.ThermocyclerStep],
        repetitions: int,
        volume: Optional[float] = None,
    ) -> None:
        """
        Begin a set temperature cycle.

        Args:
            steps: The set temperature steps.
            repetitions: Number of repetitions.
            volume: Optional volume.

        Returns: None
        """
        await self.wait_for_is_running()
        self._total_cycle_count = repetitions
        self._total_step_count = len(steps)

        task = self._loop.create_task(self._execute_cycles(steps, repetitions, volume))
        self.make_cancellable(task)
        await task

    async def execute_profile(
        self,
        profile: List[Union[types.ThermocyclerCycle, types.ThermocyclerStep]],
        volume: Optional[float] = None,
    ) -> None:
        """Begin a set temperature profile, with both repeating and non-repeating steps.

        Args:
            profile: The temperature profile to follow.
            volume: Optional volume

        Returns: None
        """
        await self.wait_for_is_running()
        self._total_cycle_count = 0
        self._total_step_count = 0
        self._current_cycle_index = 0
        self._current_step_index = 0
        for step_or_cycle in profile:
            if "steps" in step_or_cycle:
                # basically https://github.com/python/mypy/issues/14766
                this_cycle = cast(types.ThermocyclerCycle, step_or_cycle)
                self._total_cycle_count += this_cycle["repetitions"]
                self._total_step_count += (
                    len(this_cycle["steps"]) * this_cycle["repetitions"]
                )
            else:
                self._total_step_count += 1
                self._total_cycle_count += 1
        task = self._loop.create_task(self._execute_profile(profile, volume))
        self.make_cancellable(task)
        await task

    async def set_lid_temperature(self, temperature: float) -> None:
        """Set the lid temperature in degrees Celsius"""
        await self.wait_for_is_running()
        await self._driver.set_lid_temperature(temp=temperature)

        task = self._loop.create_task(self._wait_for_lid_target())
        self.make_cancellable(task)
        await task

    async def wait_for_lid_target(self) -> None:
        """Set the lid temperature in degrees Celsius"""
        await self.wait_for_is_running()

        task = self._loop.create_task(self._wait_for_lid_target())
        self.make_cancellable(task)
        await task

    # TODO(mc, 2022-04-25): de-duplicate with `set_temperature`
    async def set_target_block_temperature(
        self,
        celsius: float,
        hold_time_seconds: Optional[float] = None,
        volume: Optional[float] = None,
    ) -> None:
        """Set the Thermocycler's target block temperature.

        Does not wait for the target temperature to be reached.

        Args:
            celsius: The target block temperature, in degrees celsius.
        """
        await self.wait_for_is_running()
        await self._driver.set_plate_temperature(
            temp=celsius,
            hold_time=hold_time_seconds,
            volume=volume,
        )
        await self._reader.read_block_temperature()

    # TODO(mc, 2022-04-26): de-duplicate with `set_lid_temperature`
    async def set_target_lid_temperature(self, celsius: float) -> None:
        """Set the Thermocycler's target lid temperature.

        Does not wait for the target temperature to be reached.

        Args:
            celsius: The target lid temperature, in degrees celsius.
        """
        await self.wait_for_is_running()
        await self._driver.set_lid_temperature(temp=celsius)
        await self._reader.read_lid_temperature()

    async def _wait_for_lid_target(self) -> None:
        """
        This method only exits if lid target temperature has been reached.

        Subject to change without a version bump.
        """
        await self._reader.read_lid_temperature()

        while not _temperature_is_holding(self.lid_temp_status):
            await self._poller.wait_next_poll()

    async def _wait_for_block_target(self) -> None:
        """
        This method only exits if set temperature has been reached.

        Subject to change without a version bump.
        """
        await self._reader.read_block_temperature()

        while not _temperature_is_holding(self.status):
            await self._poller.wait_next_poll()

        while self.hold_time is not None and self.hold_time > 0:
            await self._poller.wait_next_poll()

    async def _wait_for_lid_status(self, status: ThermocyclerLidStatus) -> None:
        """Wait for lid status to be status."""
        await self._reader.read_lid_status()

        while self.lid_status != status:
            await self._poller.wait_next_poll()

    @property
    def lid_target(self) -> Optional[float]:
        return self._reader.lid_temperature.target

    @property
    def lid_temp(self) -> float:
        return self._reader.lid_temperature.current

    @property
    def lid_status(self) -> ThermocyclerLidStatus:
        return self._reader.lid_status

    @property
    def lid_temp_status(self) -> TemperatureStatus:
        return self._reader.lid_temperature_status

    # TODO(mc, 2022-10-13): remove
    @property
    def ramp_rate(self) -> Optional[float]:
        """Not supported."""
        return None

    @property
    def hold_time(self) -> Optional[float]:
        return self._reader.block_temperature.hold

    @property
    def temperature(self) -> Optional[float]:
        return self._reader.block_temperature.current

    @property
    def target(self) -> Optional[float]:
        return self._reader.block_temperature.target

    @property
    def status(self) -> TemperatureStatus:
        return (
            self._reader.block_temperature_status
            if self._error is None
            else TemperatureStatus.ERROR
        )

    @property
    def device_info(self) -> Mapping[str, str]:
        return self._device_info

    @property
    def total_cycle_count(self) -> Optional[int]:
        return self._total_cycle_count

    @property
    def current_cycle_index(self) -> Optional[int]:
        return self._current_cycle_index

    @property
    def total_step_count(self) -> Optional[int]:
        return self._total_step_count

    @property
    def current_step_index(self) -> Optional[int]:
        return self._current_step_index

    @property
    def live_data(self) -> types.LiveData:
        data: types.ThermocyclerData = {
            "lid": self.lid_status,
            "lidTarget": self.lid_target,
            "lidTemp": self.lid_temp,
            "lidTempStatus": self.lid_temp_status,
            "currentTemp": self.temperature,
            "targetTemp": self.target,
            "holdTime": self.hold_time,
            "rampRate": self.ramp_rate,
            "currentCycleIndex": self.current_cycle_index,
            "totalCycleCount": self.total_cycle_count,
            "currentStepIndex": self.current_step_index,
            "totalStepCount": self.total_step_count,
        }
        return {
            "status": self.status,
            "data": data,
        }

    @property
    def is_simulated(self) -> bool:
        return isinstance(self._driver, SimulatingDriver)

    async def prep_for_update(self) -> str:
        await self._poller.stop()
        await self._driver.enter_programming_mode()

        if self.model() == V2_MODULE_STRING:
            # TC2 has three unique "devices" over DFU
            new_port = await update.find_dfu_device(
                pid=DFU_PID, expected_device_count=3
            )
        else:
            new_port = await update.find_bootloader_port()

        return new_port or self.port

    async def _execute_cycle_step(
        self, step: types.ThermocyclerStep, volume: Optional[float]
    ) -> None:
        """
        Execute a thermocycler step.

        Args:
            step: The set temperature parameters.
            volume: The volume

        Returns: None
        """
        temperature = step.get("temperature")
        hold_time_minutes = step.get("hold_time_minutes", None)
        hold_time_seconds = step.get("hold_time_seconds", None)
        await self._set_temperature_no_pause(
            temperature=temperature,  # type: ignore
            hold_time_minutes=hold_time_minutes,
            hold_time_seconds=hold_time_seconds,
            ramp_rate=None,
            volume=volume,
        )

    async def _execute_cycles(
        self,
        steps: List[types.ThermocyclerStep],
        repetitions: int,
        volume: Optional[float],
    ) -> None:
        """
        Execute cycles.

        Args:
            steps: The set temperature steps.
            repetitions: The number of repetitions
            volume: The optional volume.

        Returns: None
        """
        for rep in range(repetitions):
            self._current_cycle_index = rep + 1  # science starts at 1
            for step_idx, step in enumerate(steps):
                self._current_step_index = step_idx + 1  # science starts at 1
                await self._execute_cycle_step(step, volume)

    async def _execute_profile(
        self,
        profile: List[Union[types.ThermocyclerCycle, types.ThermocyclerStep]],
        volume: Optional[float],
    ) -> None:
        """
        Execute profiles.

        Profiles command a thermocycler pattern that can contain multiple cycles and out-of-cycle steps.
        """
        self._current_cycle_index = 0
        self._current_step_index = 0
        for step_or_cycle in profile:
            self._current_cycle_index += 1
            if "repetitions" in step_or_cycle:
                # basically https://github.com/python/mypy/issues/14766
                this_cycle = cast(types.ThermocyclerCycle, step_or_cycle)
                for rep in range(this_cycle["repetitions"]):
                    for step in this_cycle["steps"]:
                        self._current_step_index += 1
                        await self._execute_cycle_step(step, volume)
            else:
                await self._execute_cycle_step(step_or_cycle, volume)

    # TODO(mc, 2022-10-13): why does this exist?
    # Do the driver and poller really need to be disconnected?
    # Could we accomplish the same thing by latching the error state
    # and allowing the driver and poller to continue?
    def _enter_error_state(self, error: Exception) -> None:
        """Enter into an error state.

        The Thermocycler will not be accessible in this state.

        Args:
            cause: The cause of the exception.

        Returns:
            None
        """
        self._error = str(error)
        log.error(
            f"Thermocycler has encountered an unrecoverable error: {self._error}. "
            f"Please refer to support article at "
            f"https://support.opentrons.com/en/articles/3469797-thermocycler-module"
            f" for troubleshooting."
        )
        asyncio.run_coroutine_threadsafe(self.cleanup(), self._loop)


class ThermocyclerReader(Reader):
    """Read data from the thermocycler.

    Args:
        driver: A connected Thermocycler driver.
    """

    lid_status: ThermocyclerLidStatus
    lid_temperature: Temperature
    block_temperature: PlateTemperature

    def __init__(
        self,
        driver: AbstractThermocyclerDriver,
    ) -> None:
        self.lid_status = ThermocyclerLidStatus.UNKNOWN
        self.lid_temperature = Temperature(current=25.0, target=None)
        self.block_temperature = PlateTemperature(current=25.0, target=None, hold=None)
        self._lid_temperature_status = LidTemperatureStatus()
        self._block_temperature_status = PlateTemperatureStatus()
        self._driver = driver
        self._handle_error: Optional[Callable[[Exception], None]] = None

    @property
    def block_temperature_status(self) -> TemperatureStatus:
        return self._block_temperature_status.status

    @property
    def lid_temperature_status(self) -> TemperatureStatus:
        return self._lid_temperature_status.status

    def on_error(self, exception: Exception) -> None:
        if self._handle_error is not None:
            self._handle_error(exception)

    def register_error_handler(self, handle_error: Callable[[Exception], None]) -> None:
        self._handle_error = handle_error

    async def read(self) -> None:
        """Poll the thermocycler."""
        await self.read_lid_status()
        await self.read_lid_temperature()
        await self.read_block_temperature()

    async def read_lid_status(self) -> None:
        self.lid_status = await self._driver.get_lid_status()

    async def read_block_temperature(self) -> None:
        self.block_temperature = await self._driver.get_plate_temperature()
        self._block_temperature_status.update(self.block_temperature)

    async def read_lid_temperature(self) -> None:
        self.lid_temperature = await self._driver.get_lid_temperature()
        self._lid_temperature_status.update(self.lid_temperature)



"""Pressure Dispense-in-place command request, result, and implementation models."""

from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Type, Union
from typing_extensions import Literal

from .pipetting_common import (
    PipetteIdMixin,
    FlowRateMixin,
    DispenseVolumeMixin,
    BaseLiquidHandlingResult,
    dispense_in_place,
    increase_evo_disp_count,
    DEFAULT_CORRECTION_VOLUME,
)
from .movement_common import (
    LiquidHandlingWellLocationMixin,
    StallOrCollisionError,
    move_to_well,
)

from .command import (
    AbstractCommandImpl,
    BaseCommand,
    BaseCommandCreate,
    SuccessData,
    DefinedErrorData,
)
from ..state.update_types import StateUpdate
from ..errors import ProtocolEngineError

if TYPE_CHECKING:
    from ..execution import PipettingHandler, GantryMover, MovementHandler
    from ..resources import ModelUtils
    from ..state.state import StateView


PressureDispenseCommandType = Literal["pressureDispense"]


class PressureDispenseParams(
    PipetteIdMixin, DispenseVolumeMixin, FlowRateMixin, LiquidHandlingWellLocationMixin
):
    """Payload required to pressure dispense in place."""

    pass


class PressureDispenseResult(BaseLiquidHandlingResult):
    """Result data from the execution of a PressureDispense command."""

    pass


_ExecuteReturn = Union[
    SuccessData[PressureDispenseResult],
    DefinedErrorData[StallOrCollisionError],
]


class PressureDispenseImplementation(
    AbstractCommandImpl[PressureDispenseParams, _ExecuteReturn]
):
    """Pressure dispense command implementation."""

    def __init__(
        self,
        pipetting: PipettingHandler,
        state_view: StateView,
        gantry_mover: GantryMover,
        model_utils: ModelUtils,
        movement: MovementHandler,
        **kwargs: object,
    ) -> None:
        self._pipetting = pipetting
        self._state_view = state_view
        self._gantry_mover = gantry_mover
        self._model_utils = model_utils
        self._movement = movement

    async def execute(self, params: PressureDispenseParams) -> _ExecuteReturn:
        """Move to and pressure dispense to the requested well."""
        well_location = params.wellLocation
        labware_id = params.labwareId
        well_name = params.wellName

        move_result = await move_to_well(
            movement=self._movement,
            model_utils=self._model_utils,
            pipette_id=params.pipetteId,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        )
        if isinstance(move_result, DefinedErrorData):
            return move_result

        current_position = await self._gantry_mover.get_position(params.pipetteId)
        await increase_evo_disp_count(
            pipette_id=params.pipetteId, pipetting=self._pipetting
        )
        result = await dispense_in_place(
            pipette_id=params.pipetteId,
            volume=params.volume,
            flow_rate=params.flowRate,
            push_out=None,
            location_if_error={
                "retryLocation": (
                    current_position.x,
                    current_position.y,
                    current_position.z,
                )
            },
            pipetting=self._pipetting,
            model_utils=self._model_utils,
            correction_volume=params.correctionVolume or DEFAULT_CORRECTION_VOLUME,
        )
        if isinstance(result, DefinedErrorData):
            # TODO (chb, 2025-01-29): Remove this and the OverpressureError returns once disabled for this function
            raise ProtocolEngineError(
                message="Overpressure Error during Resin Tip Dispense Command."
            )
        return SuccessData(
            public=PressureDispenseResult(volume=result.public.volume),
            state_update=StateUpdate.reduce(
                move_result.state_update, result.state_update
            ),
        )


class PressureDispense(
    BaseCommand[
        PressureDispenseParams,
        PressureDispenseResult,
        StallOrCollisionError,
    ]
):
    """PressureDispense command model."""

    commandType: PressureDispenseCommandType = "pressureDispense"
    params: PressureDispenseParams
    result: Optional[PressureDispenseResult] = None

    _ImplementationCls: Type[
        PressureDispenseImplementation
    ] = PressureDispenseImplementation


class PressureDispenseCreate(BaseCommandCreate[PressureDispenseParams]):
    """PressureDispense command request model."""

    commandType: PressureDispenseCommandType = "pressureDispense"
    params: PressureDispenseParams

    _CommandCls: Type[PressureDispense] = PressureDispense



"""Top-level ProtocolEngine configuration options."""
from dataclasses import dataclass

from opentrons_shared_data.robot.types import RobotType

from opentrons.protocol_engine.types import DeckType


@dataclass(frozen=True)
class Config:
    """ProtocolEngine configuration options.

    Params:
        robot_type: What kind of robot the engine is controlling,
            or pretending to control.
        deck_type: The type of deck on the robot that the engine is controlling,
            or pretending to control.
        ignore_pause: The engine should no-op instead of waiting
            for pauses and delays to complete.
        use_virtual_pipettes: The engine should no-op instead of calling
            instruments' hardware control API
        use_virtual_modules: The engine should no-op instead of calling
            modules' hardware control API.
        use_virtual_gripper: The engine should no-op instead of calling
            gripper hardware control API.
        use_simulated_deck_config: The engine should lazily populate the deck
            configuration instead of loading a provided configuration
        block_on_door_open: Protocol execution should pause if the
            front door is opened.
    """

    robot_type: RobotType
    deck_type: DeckType
    ignore_pause: bool = False
    use_virtual_pipettes: bool = False
    use_virtual_modules: bool = False
    use_virtual_gripper: bool = False
    use_simulated_deck_config: bool = False
    block_on_door_open: bool = False



import json
import typing
import logging
from pydantic import ValidationError
from dataclasses import asdict

from opentrons import config, types

from .. import file_operators as io, types as local_types

from .models import v1

from opentrons.types import Mount, Point
from opentrons.util.helpers import utc_now

log = logging.getLogger(__name__)

# Delete Pipette Offset Calibrations


def delete_pipette_offset_file(pipette: str, mount: Mount) -> None:
    """
    Delete pipette offset file based on mount and pipette serial number

    :param pipette: pipette serial number
    :param mount: pipette mount
    """
    offset_dir = config.get_opentrons_path("pipette_calibration_dir")
    offset_path = offset_dir / mount.name.lower() / f"{pipette}.json"
    io.delete_file(offset_path)


def clear_pipette_offset_calibrations() -> None:
    """
    Delete all pipette offset calibration files.
    """
    io._remove_json_files_in_directories(
        config.get_opentrons_path("pipette_calibration_dir")
    )


# Save Pipette Offset Calibrations


def save_pipette_calibration(
    offset: Point,
    pip_id: str,
    mount: Mount,
    tiprack_hash: str,
    tiprack_uri: str,
    cal_status: typing.Optional[
        typing.Union[local_types.CalibrationStatus, v1.CalibrationStatus]
    ] = None,
) -> None:
    pip_dir = config.get_opentrons_path("pipette_calibration_dir") / mount.name.lower()

    if isinstance(cal_status, local_types.CalibrationStatus):
        cal_status_model = v1.CalibrationStatus(**asdict(cal_status))
    elif isinstance(cal_status, v1.CalibrationStatus):
        cal_status_model = cal_status
    else:
        cal_status_model = v1.CalibrationStatus()

    pipette_calibration = v1.InstrumentOffsetModel(
        offset=offset,
        tiprack=tiprack_hash,
        uri=tiprack_uri,
        last_modified=utc_now(),
        source=local_types.SourceType.user,
        status=cal_status_model,
    )
    io.save_to_file(pip_dir, pip_id, pipette_calibration)


# Get Pipette Offset Calibrations


def get_pipette_offset(
    pipette_id: str, mount: Mount
) -> typing.Optional[v1.InstrumentOffsetModel]:
    try:
        pipette_calibration_filepath = (
            config.get_opentrons_path("pipette_calibration_dir")
            / mount.name.lower()
            / f"{pipette_id}.json"
        )
        return v1.InstrumentOffsetModel(
            **io.read_cal_file(pipette_calibration_filepath)
        )
    except FileNotFoundError:
        log.debug(f"Calibrations for {pipette_id} on {mount} does not exist.")
        return None
    except (json.JSONDecodeError, ValidationError):
        log.warning(
            f"Malformed calibrations for {pipette_id} on {mount}. Please factory reset your calibrations.",
            exc_info=True,
        )
        # TODO: Delete the bad calibration here maybe?
        return None


def get_all_pipette_offset_calibrations() -> typing.List[v1.PipetteOffsetCalibration]:
    """
    A helper function that will list all of the pipette offset
    calibrations.

    :return: A list of dictionary objects representing all of the
    pipette offset calibration files found on the robot.
    """
    pipette_calibration_dir = config.get_opentrons_path("pipette_calibration_dir")
    pipette_calibration_list = []
    for filepath in pipette_calibration_dir.glob("**/*.json"):
        pipette_id = filepath.stem
        mount = Mount.string_to_mount(filepath.parent.stem)
        calibration = get_pipette_offset(pipette_id, mount)
        if calibration:
            pipette_calibration_list.append(
                v1.PipetteOffsetCalibration(
                    pipette=pipette_id,
                    mount=mount.name.lower(),
                    offset=types.Point(*calibration.offset),
                    tiprack=calibration.tiprack,
                    uri=calibration.uri,
                    last_modified=calibration.last_modified,
                    source=calibration.source,
                    status=calibration.status,
                )
            )
    return pipette_calibration_list



"""Labware movement command handling."""

from __future__ import annotations

from typing import Optional, TYPE_CHECKING, overload

from opentrons_shared_data.labware.labware_definition import LabwareDefinition

from opentrons.types import Point

from opentrons.hardware_control import HardwareControlAPI
from opentrons.hardware_control.types import OT3Mount, Axis
from opentrons.motion_planning import get_gripper_labware_movement_waypoints

from opentrons.protocol_engine.state.state import StateStore
from opentrons.protocol_engine.resources.ot3_validation import ensure_ot3_hardware

from .thermocycler_movement_flagger import ThermocyclerMovementFlagger
from .heater_shaker_movement_flagger import HeaterShakerMovementFlagger
from .thermocycler_plate_lifter import ThermocyclerPlateLifter

from ..errors import (
    GripperNotAttachedError,
    LabwareMovementNotAllowedError,
    ThermocyclerNotOpenError,
    HeaterShakerLabwareLatchNotOpenError,
    CannotPerformGripperAction,
)

from ..types import (
    OnLabwareLocation,
    LabwareLocation,
    LabwareMovementOffsetData,
    OnDeckLabwareLocation,
)

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import EquipmentHandler, MovementHandler

_GRIPPER_HOMED_POSITION_Z = 166.125  # Height of the center of the gripper critical point from the deck when homed


class LabwareMovementHandler:
    """Implementation logic for labware movement."""

    _hardware_api: HardwareControlAPI
    _state_store: StateStore
    _movement: MovementHandler
    _equipment: EquipmentHandler

    def __init__(
        self,
        hardware_api: HardwareControlAPI,
        state_store: StateStore,
        equipment: EquipmentHandler,
        movement: MovementHandler,
        thermocycler_plate_lifter: Optional[ThermocyclerPlateLifter] = None,
        thermocycler_movement_flagger: Optional[ThermocyclerMovementFlagger] = None,
        heater_shaker_movement_flagger: Optional[HeaterShakerMovementFlagger] = None,
    ) -> None:
        """Initialize a LabwareMovementHandler instance."""
        self._hardware_api = hardware_api
        self._state_store = state_store
        self._equipment = equipment
        self._thermocycler_plate_lifter = (
            thermocycler_plate_lifter
            or ThermocyclerPlateLifter(
                state_store=self._state_store,
                equipment=equipment,
                movement=movement,
            )
        )
        self._tc_movement_flagger = (
            thermocycler_movement_flagger
            or ThermocyclerMovementFlagger(
                state_store=self._state_store,
                hardware_api=self._hardware_api,
                equipment=self._equipment
                or EquipmentHandler(
                    hardware_api=self._hardware_api,
                    state_store=self._state_store,
                ),
            )
        )
        self._hs_movement_flagger = (
            heater_shaker_movement_flagger
            or HeaterShakerMovementFlagger(
                state_store=self._state_store, hardware_api=self._hardware_api
            )
        )

    @overload
    async def move_labware_with_gripper(
        self,
        *,
        labware_id: str,
        current_location: OnDeckLabwareLocation,
        new_location: OnDeckLabwareLocation,
        user_offset_data: LabwareMovementOffsetData,
        post_drop_slide_offset: Optional[Point],
    ) -> None:
        ...

    @overload
    async def move_labware_with_gripper(
        self,
        *,
        labware_definition: LabwareDefinition,
        current_location: OnDeckLabwareLocation,
        new_location: OnDeckLabwareLocation,
        user_offset_data: LabwareMovementOffsetData,
        post_drop_slide_offset: Optional[Point],
    ) -> None:
        ...

    async def move_labware_with_gripper(  # noqa: C901
        self,
        *,
        labware_id: str | None = None,
        labware_definition: LabwareDefinition | None = None,
        current_location: OnDeckLabwareLocation,
        new_location: OnDeckLabwareLocation,
        user_offset_data: LabwareMovementOffsetData,
        post_drop_slide_offset: Optional[Point],
    ) -> None:
        """Physically move a labware from one location to another using the gripper.

        Generally, provide the `labware_id` of a loaded labware, and this method will
        automatically look up its labware definition. If you're physically moving
        something that has not been loaded as a labware (this is not common),
        provide the `labware_definition` yourself instead.
        """
        use_virtual_gripper = self._state_store.config.use_virtual_gripper

        if labware_definition is None:
            assert labware_id is not None  # From this method's @typing.overloads.
            labware_definition = self._state_store.labware.get_definition(labware_id)

        from_labware_center = self._state_store.geometry.get_labware_grip_point(
            labware_definition=labware_definition, location=current_location
        )
        to_labware_center = self._state_store.geometry.get_labware_grip_point(
            labware_definition=labware_definition, location=new_location
        )

        if use_virtual_gripper:
            # todo(mm, 2024-11-07): We should do this collision checking even when we
            # only have a `labware_definition`, not a `labware_id`. Resolve when
            # `check_gripper_labware_tip_collision()` can be made independent of `labware_id`.
            if labware_id is not None:
                self._state_store.geometry.check_gripper_labware_tip_collision(
                    # During Analysis we will pass in hard coded estimates for certain positions only accessible during execution
                    gripper_homed_position_z=_GRIPPER_HOMED_POSITION_Z,
                    labware_id=labware_id,
                    current_location=current_location,
                )
            return

        ot3api = ensure_ot3_hardware(
            hardware_api=self._hardware_api,
            error_msg="Gripper is only available on Opentrons Flex",
        )

        if not ot3api.has_gripper():
            raise GripperNotAttachedError(
                "No gripper found for performing labware movements."
            )
        if not ot3api.gripper_jaw_can_home():
            raise CannotPerformGripperAction(
                "Cannot pick up labware when gripper is already gripping."
            )

        gripper_mount = OT3Mount.GRIPPER

        # Retract all mounts
        await ot3api.home(axes=[Axis.Z_L, Axis.Z_R, Axis.Z_G])
        gripper_homed_position = await ot3api.gantry_position(mount=gripper_mount)

        # todo(mm, 2024-11-07): We should do this collision checking even when we
        # only have a `labware_definition`, not a `labware_id`. Resolve when
        # `check_gripper_labware_tip_collision()` can be made independent of `labware_id`.
        if labware_id is not None:
            self._state_store.geometry.check_gripper_labware_tip_collision(
                gripper_homed_position_z=gripper_homed_position.z,
                labware_id=labware_id,
                current_location=current_location,
            )

        async with self._thermocycler_plate_lifter.lift_plate_for_labware_movement(
            labware_location=current_location
        ):
            final_offsets = (
                self._state_store.geometry.get_final_labware_movement_offset_vectors(
                    from_location=current_location,
                    to_location=new_location,
                    additional_offset_vector=user_offset_data,
                    current_labware=labware_definition,
                )
            )
            movement_waypoints = get_gripper_labware_movement_waypoints(
                from_labware_center=from_labware_center,
                to_labware_center=to_labware_center,
                gripper_home_z=gripper_homed_position.z,
                offset_data=final_offsets,
                post_drop_slide_offset=post_drop_slide_offset,
            )
            labware_grip_force = self._state_store.labware.get_grip_force(
                labware_definition
            )
            holding_labware = False
            for waypoint_data in movement_waypoints:
                if waypoint_data.jaw_open:
                    if waypoint_data.dropping:
                        # This `disengage_axes` step is important in order to engage
                        # the electronic brake on the Z axis of the gripper. The brake
                        # has a stronger holding force on the axis than the hold current,
                        # and prevents the axis from spuriously dropping when  e.g. the notch
                        # on the side of a falling tiprack catches the jaw.
                        await ot3api.disengage_axes([Axis.Z_G])
                    await ot3api.ungrip()
                    holding_labware = True
                    if waypoint_data.dropping:
                        # We lost the position estimation after disengaging the axis, so
                        # it is necessary to home it next
                        await ot3api.home_z(OT3Mount.GRIPPER)
                else:
                    await ot3api.grip(force_newtons=labware_grip_force)
                    # we only want to check position after the gripper has opened and
                    # should be holding labware
                    if holding_labware:
                        labware_bbox = self._state_store.labware.get_dimensions(
                            labware_definition=labware_definition
                        )
                        well_bbox = self._state_store.labware.get_well_bbox(
                            labware_definition=labware_definition
                        )
                        # todo(mm, 2024-09-26): This currently raises a lower-level 2015 FailedGripperPickupError.
                        # Convert this to a higher-level 3001 LabwareDroppedError or 3002 LabwareNotPickedUpError,
                        # depending on what waypoint we're at, to propagate a more specific error code to users.
                        ot3api.raise_error_if_gripper_pickup_failed(
                            expected_grip_width=labware_bbox.y,
                            grip_width_uncertainty_wider=abs(
                                max(well_bbox.y - labware_bbox.y, 0)
                            ),
                            grip_width_uncertainty_narrower=abs(
                                min(well_bbox.y - labware_bbox.y, 0)
                            ),
                        )
                await ot3api.move_to(
                    mount=gripper_mount, abs_position=waypoint_data.position
                )

            # this makes sure gripper jaw is closed between two move labware calls
            await ot3api.idle_gripper()

    async def ensure_movement_not_obstructed_by_module(
        self, labware_id: str, new_location: LabwareLocation
    ) -> None:
        """Ensure that the labware movement is not obstructed by a parent module.

        Raises: LabwareMovementNotAllowedError if either current location or
        new location is a module that is in a state that prevents the labware from
        being moved (either manually or using gripper).
        """
        current_parent = self._state_store.labware.get_parent_location(
            labware_id=labware_id
        )
        if isinstance(new_location, OnLabwareLocation):
            new_location = self._state_store.labware.get_parent_location(
                labware_id=new_location.labwareId
            )
        for parent in (current_parent, new_location):
            try:
                await self._tc_movement_flagger.ensure_labware_in_open_thermocycler(
                    labware_parent=parent
                )
                await self._hs_movement_flagger.raise_if_labware_latched_on_heater_shaker(
                    labware_parent=parent
                )
            except ThermocyclerNotOpenError:
                raise LabwareMovementNotAllowedError(
                    "Cannot move labware to or from a Thermocycler with its lid closed."
                )
            except HeaterShakerLabwareLatchNotOpenError:
                raise LabwareMovementNotAllowedError(
                    "Cannot move labware to or from a Heater-Shaker"
                    " with its labware latch closed."
                )



import logging
from functools import partial
from typing import cast, Callable, Optional, List, Set
from typing_extensions import TypedDict, Literal

from opentrons.hardware_control.types import TipStateType, OT3Mount, InstrumentProbeType

from opentrons_hardware.drivers.can_bus import CanMessenger
from opentrons_hardware.firmware_bindings.constants import NodeId
from opentrons_hardware.hardware_control.tip_presence import (
    TipDetector,
    types as tip_types,
)
from opentrons_shared_data.errors.exceptions import (
    TipDetectorNotFound,
    UnmatchedTipPresenceStates,
    GeneralError,
)

from .ot3utils import sensor_id_for_instrument

log = logging.getLogger(__name__)

TipListener = Callable[[OT3Mount, bool], None]
PipetteMountKeys = Literal["left", "right"]


class TipDetectorByMount(TypedDict):
    left: Optional[TipDetector]
    right: Optional[TipDetector]


class UnsubMethodByMount(TypedDict):
    left: Optional[Callable[[], None]]
    right: Optional[Callable[[], None]]


class TipUpdateByMount(TypedDict):
    left: Optional[bool]
    right: Optional[bool]


def _mount_to_node(mount: OT3Mount) -> NodeId:
    return {
        OT3Mount.LEFT: NodeId.pipette_left,
        OT3Mount.RIGHT: NodeId.pipette_right,
    }[mount]


class TipPresenceManager:
    """Handle tip change notification coming from CAN."""

    _listeners: Set[TipListener]
    _detectors: TipDetectorByMount
    _unsub_methods: UnsubMethodByMount
    _last_state: TipUpdateByMount

    def __init__(
        self,
        can_messenger: CanMessenger,
        listeners: Set[TipListener] = set(),
    ) -> None:
        self._messenger = can_messenger
        self._listeners = listeners
        self._detectors = TipDetectorByMount(left=None, right=None)
        self._unsub_methods = UnsubMethodByMount(left=None, right=None)
        self._last_state = TipUpdateByMount(left=None, right=None)

    @staticmethod
    def _get_key(mount: OT3Mount) -> PipetteMountKeys:
        assert mount != OT3Mount.GRIPPER
        return cast(PipetteMountKeys, mount.name.lower())

    async def clear_detector(self, mount: OT3Mount) -> None:
        """Clean up and remove tip detector."""

        def _unsubscribe() -> None:
            """Unsubscribe from detector."""
            unsub = self._unsub_methods[self._get_key(mount)]
            if unsub:
                unsub()
                self.set_unsub(mount, None)

        try:
            detector = self.get_detector(mount)
            detector.cleanup()
        except TipDetectorNotFound:
            pass
        finally:
            _unsubscribe()
            self.set_detector(mount, None)

    async def build_detector(self, mount: OT3Mount, sensor_count: int) -> None:
        assert self._detectors.get(self._get_key(mount), None) is None
        # set up and subscribe to the detector
        d = TipDetector(self._messenger, _mount_to_node(mount), sensor_count)
        # listens to the detector so we can immediately notify listeners
        # the most up-to-date tip state
        unsub = d.add_subscriber(partial(self._handle_tip_update, mount))
        self.set_unsub(mount, unsub)
        self.set_detector(mount, d)

    def _handle_tip_update(
        self, mount: OT3Mount, update: tip_types.TipNotification
    ) -> None:
        """Callback for detector."""
        self._last_state[self._get_key(mount)] = update.presence

        for listener in self._listeners:
            listener(mount, update.presence)

    def current_tip_state(self, mount: OT3Mount) -> Optional[bool]:
        state = self._last_state[self._get_key(mount)]
        return state

    @staticmethod
    def _get_tip_presence(
        results: List[tip_types.TipNotification],
        follow_singular_sensor: Optional[InstrumentProbeType] = None,
    ) -> TipStateType:
        """
        We can use follow_singular_sensor used to specify that we only care
        about the status of one tip presence sensor on a high throughput
        pipette, and the other is allowed to be different.
        """
        if follow_singular_sensor:
            target_sensor_id = sensor_id_for_instrument(follow_singular_sensor)
            for r in results:
                if r.sensor == target_sensor_id:
                    return TipStateType(r.presence)
            # raise an error if requested sensor response isn't found
            raise GeneralError(
                message=f"Requested status for sensor {follow_singular_sensor} not found."
            )
        # more than one sensor reported, we have to check if their states match
        if len(set(r.presence for r in results)) > 1:
            raise UnmatchedTipPresenceStates(
                {int(r.sensor): int(r.presence) for r in results}
            )
        return TipStateType(results[0].presence)

    async def get_tip_status(
        self,
        mount: OT3Mount,
        follow_singular_sensor: Optional[InstrumentProbeType] = None,
    ) -> TipStateType:
        detector = self.get_detector(mount)
        return self._get_tip_presence(
            await detector.request_tip_status(), follow_singular_sensor
        )

    def get_detector(self, mount: OT3Mount) -> TipDetector:
        detector = self._detectors[self._get_key(mount)]
        if not detector:
            raise TipDetectorNotFound(
                message=f"Tip detector not set up for {mount} mount",
                detail={"mount": str(mount)},
            )
        return detector

    def set_detector(self, mount: OT3Mount, detector: Optional[TipDetector]) -> None:
        self._detectors[self._get_key(mount)] = detector

    def set_unsub(self, mount: OT3Mount, unsub: Optional[Callable[[], None]]) -> None:
        self._unsub_methods[self._get_key(mount)] = unsub

    def add_listener(self, listener: TipListener) -> Callable[[], None]:
        self._listeners.add(listener)

        def remove() -> None:
            self._listeners.discard(listener)

        return remove



"""Simple state change notification interface."""
import asyncio


class ChangeNotifier:
    """An interface to emit or subscribe to state change notifications."""

    def __init__(self) -> None:
        """Initialize the ChangeNotifier with an internal Event."""
        self._event = asyncio.Event()

    def notify(self) -> None:
        """Notify all `wait`'ers that the state has changed."""
        self._event.set()

    async def wait(self) -> None:
        """Wait until the next state change notification."""
        await self._event.wait()
        self._event.clear()


class ChangeNotifier_ts(ChangeNotifier):
    """ChangeNotifier initialized with Event_ts."""

    def __init__(self) -> None:
        """Initialize the ChangeNotifier_Ts with an internal Event_ts."""
        super().__init__()
        self._event = Event_ts()


class Event_ts(asyncio.Event):
    """asyncio.Event with threadsafe methods."""

    def __init__(self) -> None:
        """Initialize Event_ts with the active event_loop or event_loop_policy if not active."""
        super().__init__()
        if self._loop is None:
            self._loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()

    def set(self) -> None:
        """Primarily intended for calling from a thread not responsible for the event loop.

        Calling set() from the event loop thread will actually delay the execution of the set() until the
        calling method either yields, awaits, or exits altogether. This is usually fine but might occasionally cause
        unexpected behavior.
        """
        self._loop.call_soon_threadsafe(super().set)



from typing import NamedTuple, Union, Optional

from opentrons.types import (
    Mount,
    DeckLocation,
    Location,
    Point,
    AxisMapType,
    AxisType,
    StringAxisMap,
)
from opentrons.legacy_broker import LegacyBroker
from opentrons.legacy_commands import robot_commands as cmds
from opentrons.legacy_commands import publisher
from opentrons.hardware_control import SyncHardwareAPI
from opentrons.protocols.api_support.util import requires_version
from opentrons.protocols.api_support.types import APIVersion
from opentrons_shared_data.pipette.types import PipetteNameType

from . import validation
from .core.common import ProtocolCore, RobotCore
from .module_contexts import ModuleContext
from .labware import Labware
from ._types import PipetteActionTypes, PlungerPositionTypes


class HardwareManager(NamedTuple):
    """Back. compat. wrapper for a removed class called `HardwareManager`.

    This interface will not be present in PAPIv3.
    """

    hardware: SyncHardwareAPI


class RobotContext(publisher.CommandPublisher):
    """
    A context for the movement system of the robot.

    The RobotContext class provides the objects, attributes, and methods that allow
    you to control robot motor axes individually.

    Its methods can command the robot to perform an action, like moving to an absolute position,
    controlling the gripper jaw, or moving individual pipette motors.

    Objects in this class should not be instantiated directly. Instead, instances are
    returned by :py:meth:`ProtocolContext.robot`.

    .. versionadded:: 2.20

    """

    def __init__(
        self,
        core: RobotCore,
        protocol_core: ProtocolCore,
        api_version: APIVersion,
        broker: Optional[LegacyBroker] = None,
    ) -> None:
        super().__init__(broker)
        self._hardware = HardwareManager(hardware=protocol_core.get_hardware())
        self._core = core
        self._protocol_core = protocol_core
        self._api_version = api_version

    @property
    @requires_version(2, 22)
    def api_version(self) -> APIVersion:
        return self._api_version

    @property
    def hardware(self) -> HardwareManager:
        # TODO this hardware attribute should be deprecated
        # in version 3.0+ as we will only support exposed robot
        # context commands.
        return self._hardware

    @requires_version(2, 22)
    def move_to(
        self,
        mount: Union[Mount, str],
        destination: Location,
        speed: Optional[float] = None,
    ) -> None:
        """
        Move a specified mount to a destination location on the deck.

        :param mount: The mount of the instrument you wish to move.
                      This can either be an instance of :py:class:`.types.Mount` or one
                      of the strings ``"left"``, ``"right"``, ``"extension"``, ``"gripper"``. Note
                      that the gripper mount can be referred to either as ``"extension"`` or ``"gripper"``.
        :type mount: types.Mount or str
        :param Location destination:
        :param speed:
        """
        mount = validation.ensure_instrument_mount(mount)
        with publisher.publish_context(
            broker=self.broker,
            command=cmds.move_to(
                # This needs to be called from protocol context and not the command for import loop reasons
                mount=mount,
                location=destination,
                speed=speed,
            ),
        ):
            self._core.move_to(mount, destination.point, speed)

    @requires_version(2, 22)
    def move_axes_to(
        self,
        axis_map: Union[AxisMapType, StringAxisMap],
        critical_point: Optional[Union[AxisMapType, StringAxisMap]] = None,
        speed: Optional[float] = None,
    ) -> None:
        """
        Move a set of axes to an absolute position on the deck.

        :param axis_map: A dictionary mapping axes to an absolute position on the deck in mm.
        :param critical_point: The critical point to move the axes with. It should only
        specify the gantry axes (i.e. `x`, `y`, `z`).
        :param float speed: The maximum speed with which you want to move all the axes
        in the axis map.
        """
        instrument_on_left = self._core.get_pipette_type_from_engine(Mount.LEFT)
        is_96_channel = instrument_on_left == PipetteNameType.P1000_96
        axis_map = validation.ensure_axis_map_type(
            axis_map, self._protocol_core.robot_type, is_96_channel
        )
        if critical_point:
            critical_point = validation.ensure_axis_map_type(
                critical_point, self._protocol_core.robot_type, is_96_channel
            )
            validation.ensure_only_gantry_axis_map_type(
                critical_point, self._protocol_core.robot_type
            )
        else:
            critical_point = None
        with publisher.publish_context(
            broker=self.broker,
            command=cmds.move_axis_to(
                # This needs to be called from protocol context and not the command for import loop reasons
                axis_map=axis_map,
                speed=speed,
            ),
        ):
            self._core.move_axes_to(axis_map, critical_point, speed)

    @requires_version(2, 22)
    def move_axes_relative(
        self,
        axis_map: Union[AxisMapType, StringAxisMap],
        speed: Optional[float] = None,
    ) -> None:
        """
        Move a set of axes to a relative position on the deck.

        :param axis_map: A dictionary mapping axes to relative movements in mm.
        :type mount: types.Mount or str

        :param float speed: The maximum speed with which you want to move all the axes
        in the axis map.
        """
        instrument_on_left = self._core.get_pipette_type_from_engine(Mount.LEFT)
        is_96_channel = instrument_on_left == PipetteNameType.P1000_96

        axis_map = validation.ensure_axis_map_type(
            axis_map, self._protocol_core.robot_type, is_96_channel
        )
        with publisher.publish_context(
            broker=self.broker,
            command=cmds.move_axis_relative(
                # This needs to be called from protocol context and not the command for import loop reasons
                axis_map=axis_map,
                speed=speed,
            ),
        ):
            self._core.move_axes_relative(axis_map, speed)

    def close_gripper_jaw(self, force: Optional[float] = None) -> None:
        """Command the gripper closed with some force."""
        with publisher.publish_context(
            broker=self.broker,
            command=cmds.close_gripper(
                force=force,
            ),
        ):
            self._core.close_gripper(force)

    def open_gripper_jaw(self) -> None:
        """Command the gripper open."""
        with publisher.publish_context(
            broker=self.broker,
            command=cmds.open_gripper(),
        ):
            self._core.release_grip()

    def axis_coordinates_for(
        self,
        mount: Union[Mount, str],
        location: Union[Location, ModuleContext, DeckLocation],
    ) -> AxisMapType:
        """
        Build a :py:class:`.types.AxisMapType` from a location to be compatible with
        either :py:meth:`.RobotContext.move_axes_to` or :py:meth:`.RobotContext.move_axes_relative`.
        You must provide only one of `location`, `slot`, or `module` to build
        the axis map.

        :param mount: The mount of the instrument you wish create an axis map for.
                      This can either be an instance of :py:class:`.types.Mount` or one
                      of the strings ``"left"``, ``"right"``, ``"extension"``, ``"gripper"``. Note
                      that the gripper mount can be referred to either as ``"extension"`` or ``"gripper"``.
        :type mount: types.Mount or str
        :param location: The location to format an axis map for.
        :type location: `Well`, `ModuleContext`, `DeckLocation` or `OffDeckType`
        """
        mount = validation.ensure_instrument_mount(mount)

        mount_axis = AxisType.axis_for_mount(mount)
        if location:
            loc: Union[Point, Labware, None]
            if isinstance(location, ModuleContext):
                loc = location.labware
                if not loc:
                    raise ValueError(f"There must be a labware on {location}")
                top_of_labware = loc.wells()[0].top()
                loc = top_of_labware.point
                return {mount_axis: loc.z, AxisType.X: loc.x, AxisType.Y: loc.y}
            elif location is DeckLocation and not isinstance(location, Location):
                slot_name = validation.ensure_and_convert_deck_slot(
                    location,
                    api_version=self._api_version,
                    robot_type=self._protocol_core.robot_type,
                )
                loc = self._protocol_core.get_slot_center(slot_name)
                return {mount_axis: loc.z, AxisType.X: loc.x, AxisType.Y: loc.y}
            elif isinstance(location, Location):
                assert isinstance(location, Location)
                loc = location.point
                return {mount_axis: loc.z, AxisType.X: loc.x, AxisType.Y: loc.y}
            else:
                raise ValueError(
                    "Location parameter must be a Module, Deck Location, or Location type."
                )
        else:
            raise TypeError("You must specify a location to move to.")

    def plunger_coordinates_for_volume(
        self, mount: Union[Mount, str], volume: float, action: PipetteActionTypes
    ) -> AxisMapType:
        """
        Build a :py:class:`.types.AxisMapType` for a pipette plunger motor from volume.

        """
        pipette_name = self._core.get_pipette_type_from_engine(mount)
        if not pipette_name:
            raise ValueError(
                f"Expected a pipette to be attached to provided mount {mount}"
            )
        mount = validation.ensure_mount_for_pipette(mount, pipette_name)
        pipette_axis = AxisType.plunger_axis_for_mount(mount)

        pipette_position = self._core.get_plunger_position_from_volume(
            mount, volume, action, self._protocol_core.robot_type
        )
        return {pipette_axis: pipette_position}

    def plunger_coordinates_for_named_position(
        self, mount: Union[Mount, str], position_name: PlungerPositionTypes
    ) -> AxisMapType:
        """
        Build a :py:class:`.types.AxisMapType` for a pipette plunger motor from position_name.

        """
        pipette_name = self._core.get_pipette_type_from_engine(mount)
        if not pipette_name:
            raise ValueError(
                f"Expected a pipette to be attached to provided mount {mount}"
            )
        mount = validation.ensure_mount_for_pipette(mount, pipette_name)
        pipette_axis = AxisType.plunger_axis_for_mount(mount)
        pipette_position = self._core.get_plunger_position_from_name(
            mount, position_name
        )
        return {pipette_axis: pipette_position}

    def build_axis_map(self, axis_map: StringAxisMap) -> AxisMapType:
        """Take in a :py:class:`.types.StringAxisMap` and output a :py:class:`.types.AxisMapType`.
        A :py:class:`.types.StringAxisMap` is allowed to contain any of the following strings:
        ``"x"``, ``"y"``, "``z_l"``, "``z_r"``, "``z_g"``, ``"q"``.

        An example of a valid axis map could be:

        {"x": 1, "y": 2} or {"Z_L": 100}

        Note that capitalization does not matter.

        """
        instrument_on_left = self._core.get_pipette_type_from_engine(Mount.LEFT)
        is_96_channel = instrument_on_left == PipetteNameType.P1000_96

        return validation.ensure_axis_map_type(
            axis_map, self._protocol_core.robot_type, is_96_channel
        )



"""Protocol engine notes module."""

from .notes import (
    NoteKind,
    CommandNote,
    CommandNoteAdder,
    CommandNoteTracker,
    make_error_recovery_debug_note,
)

__all__ = [
    "NoteKind",
    "CommandNote",
    "CommandNoteAdder",
    "CommandNoteTracker",
    "make_error_recovery_debug_note",
]



"""Protocol engine types to do with modules."""

from __future__ import annotations
from dataclasses import dataclass
from enum import Enum
from typing import (
    TypeGuard,
    Literal,
    Optional,
    List,
    Dict,
    Any,
    NamedTuple,
)

from pydantic import BaseModel, Field

from opentrons.hardware_control.modules import (
    ModuleType as ModuleType,
)

from .location import DeckSlotLocation
from .labware_offset_vector import LabwareOffsetVector
from .labware_movement import LabwareMovementOffsetData


# TODO(mc, 2022-01-18): use opentrons_shared_data.module.types.ModuleModel
class ModuleModel(str, Enum):
    """All available modules' models."""

    TEMPERATURE_MODULE_V1 = "temperatureModuleV1"
    TEMPERATURE_MODULE_V2 = "temperatureModuleV2"
    MAGNETIC_MODULE_V1 = "magneticModuleV1"
    MAGNETIC_MODULE_V2 = "magneticModuleV2"
    THERMOCYCLER_MODULE_V1 = "thermocyclerModuleV1"
    THERMOCYCLER_MODULE_V2 = "thermocyclerModuleV2"
    HEATER_SHAKER_MODULE_V1 = "heaterShakerModuleV1"
    MAGNETIC_BLOCK_V1 = "magneticBlockV1"
    ABSORBANCE_READER_V1 = "absorbanceReaderV1"
    FLEX_STACKER_MODULE_V1 = "flexStackerModuleV1"

    def as_type(self) -> ModuleType:
        """Get the ModuleType of this model."""
        if ModuleModel.is_temperature_module_model(self):
            return ModuleType.TEMPERATURE
        elif ModuleModel.is_magnetic_module_model(self):
            return ModuleType.MAGNETIC
        elif ModuleModel.is_thermocycler_module_model(self):
            return ModuleType.THERMOCYCLER
        elif ModuleModel.is_heater_shaker_module_model(self):
            return ModuleType.HEATER_SHAKER
        elif ModuleModel.is_magnetic_block(self):
            return ModuleType.MAGNETIC_BLOCK
        elif ModuleModel.is_absorbance_reader(self):
            return ModuleType.ABSORBANCE_READER
        elif ModuleModel.is_flex_stacker(self):
            return ModuleType.FLEX_STACKER

        assert False, f"Invalid ModuleModel {self}"

    @classmethod
    def is_temperature_module_model(
        cls, model: ModuleModel
    ) -> TypeGuard[TemperatureModuleModel]:
        """Whether a given model is a Temperature Module."""
        return model in [cls.TEMPERATURE_MODULE_V1, cls.TEMPERATURE_MODULE_V2]

    @classmethod
    def is_magnetic_module_model(
        cls, model: ModuleModel
    ) -> TypeGuard[MagneticModuleModel]:
        """Whether a given model is a Magnetic Module."""
        return model in [cls.MAGNETIC_MODULE_V1, cls.MAGNETIC_MODULE_V2]

    @classmethod
    def is_thermocycler_module_model(
        cls, model: ModuleModel
    ) -> TypeGuard[ThermocyclerModuleModel]:
        """Whether a given model is a Thermocycler Module."""
        return model in [cls.THERMOCYCLER_MODULE_V1, cls.THERMOCYCLER_MODULE_V2]

    @classmethod
    def is_heater_shaker_module_model(
        cls, model: ModuleModel
    ) -> TypeGuard[HeaterShakerModuleModel]:
        """Whether a given model is a Heater-Shaker Module."""
        return model == cls.HEATER_SHAKER_MODULE_V1

    @classmethod
    def is_magnetic_block(cls, model: ModuleModel) -> TypeGuard[MagneticBlockModel]:
        """Whether a given model is a Magnetic block."""
        return model == cls.MAGNETIC_BLOCK_V1

    @classmethod
    def is_absorbance_reader(
        cls, model: ModuleModel
    ) -> TypeGuard[AbsorbanceReaderModel]:
        """Whether a given model is an Absorbance Plate Reader."""
        return model == cls.ABSORBANCE_READER_V1

    @classmethod
    def is_flex_stacker(cls, model: ModuleModel) -> TypeGuard[FlexStackerModuleModel]:
        """Whether a given model is a Flex Stacker.."""
        return model == cls.FLEX_STACKER_MODULE_V1


TemperatureModuleModel = Literal[
    ModuleModel.TEMPERATURE_MODULE_V1, ModuleModel.TEMPERATURE_MODULE_V2
]
MagneticModuleModel = Literal[
    ModuleModel.MAGNETIC_MODULE_V1, ModuleModel.MAGNETIC_MODULE_V2
]
ThermocyclerModuleModel = Literal[
    ModuleModel.THERMOCYCLER_MODULE_V1, ModuleModel.THERMOCYCLER_MODULE_V2
]
HeaterShakerModuleModel = Literal[ModuleModel.HEATER_SHAKER_MODULE_V1]
MagneticBlockModel = Literal[ModuleModel.MAGNETIC_BLOCK_V1]
AbsorbanceReaderModel = Literal[ModuleModel.ABSORBANCE_READER_V1]
FlexStackerModuleModel = Literal[ModuleModel.FLEX_STACKER_MODULE_V1]


class ModuleDimensions(BaseModel):
    """Dimension type for modules."""

    bareOverallHeight: float
    overLabwareHeight: float
    lidHeight: Optional[float] = None
    maxStackerFillHeight: Optional[float] = None
    maxStackerRetrievableHeight: Optional[float] = None


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class ModuleCalibrationPoint(BaseModel):
    """Calibration Point type for module definition."""

    x: float
    y: float
    z: float


# TODO(mm, 2023-04-13): Move to shared-data, so this binding can be maintained alongside the JSON
# schema that it's sourced from. We already do that for labware definitions and JSON protocols.
class ModuleDefinition(BaseModel):
    """A module definition conforming to module definition schema v3."""

    # Note: This field is misleading.
    #
    # This class only models v3 definitions ("module/schemas/3"), not v2 ("module/schemas/2").
    # labwareOffset is required to have a z-component, for example.
    #
    # When parsing from a schema v3 JSON definition into this model,
    # the definition's `"$otSharedSchema": "module/schemas/3"` field will be thrown away
    # because it has a dollar sign, which doesn't match this field.
    # Then, this field will default to "module/schemas/2", because no value was provided.
    #
    # We should fix this field once Jira RSS-221 is resolved. RSS-221 makes it difficult to fix
    # because robot-server has been storing and loading these bad fields in its database.
    otSharedSchema: str = Field("module/schemas/2", description="The current schema.")

    moduleType: ModuleType = Field(
        ...,
        description="Module type (Temperature/Magnetic/Thermocycler)",
    )

    model: ModuleModel = Field(..., description="Model name of the module")

    labwareOffset: LabwareOffsetVector = Field(
        ...,
        description="Labware offset in x, y, z.",
    )

    dimensions: ModuleDimensions = Field(..., description="Module dimension")

    calibrationPoint: ModuleCalibrationPoint = Field(
        ...,
        description="Calibration point of module.",
    )

    displayName: str = Field(..., description="Display name.")

    quirks: List[str] = Field(..., description="Module quirks")

    # In releases prior to https://github.com/Opentrons/opentrons/pull/11873 (v6.3.0),
    # the matrices in slotTransforms were 3x3.
    # After, they are 4x4, even though there was no schema version bump.
    #
    # Because old objects of this class, with the 3x3 matrices, were stored in robot-server's
    # database, this field needs to stay typed loosely enough to support both sizes.
    # We can fix this once Jira RSS-221 is resolved.
    slotTransforms: Dict[str, Any] = Field(
        ...,
        description="Dictionary of transforms for each slot.",
    )

    compatibleWith: List[ModuleModel] = Field(
        ...,
        description="List of module models this model is compatible with.",
    )
    gripperOffsets: Optional[Dict[str, LabwareMovementOffsetData]] = Field(
        default_factory=dict,
        description="Offsets to use for labware movement using gripper",
    )


class LoadedModule(BaseModel):
    """A module that has been loaded."""

    id: str
    model: ModuleModel
    location: Optional[DeckSlotLocation] = None
    serialNumber: Optional[str] = None


class SpeedRange(NamedTuple):
    """Minimum and maximum allowed speeds for a shaking module."""

    min: int
    max: int


class TemperatureRange(NamedTuple):
    """Minimum and maximum allowed temperatures for a heating module."""

    min: float
    max: float


class HeaterShakerLatchStatus(Enum):
    """Heater-Shaker latch status for determining pipette and labware movement errors."""

    CLOSED = "closed"
    OPEN = "open"
    UNKNOWN = "unknown"


@dataclass(frozen=True)
class HeaterShakerMovementRestrictors:
    """Shaking status, latch status and slot location for determining movement restrictions."""

    plate_shaking: bool
    latch_status: HeaterShakerLatchStatus
    deck_slot: int


ABSMeasureMode = Literal["single", "multi"]


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class ModuleOffsetVector(BaseModel):
    """Offset, in deck coordinates, from nominal to actual position of labware on a module."""

    x: float
    y: float
    z: float

    def __add__(self, other: Any) -> ModuleOffsetVector:
        """Adds two vectors together."""
        if not isinstance(other, (LabwareOffsetVector, ModuleOffsetVector)):
            return NotImplemented
        return ModuleOffsetVector(
            x=self.x + other.x, y=self.y + other.y, z=self.z + other.z
        )

    def __radd__(self, other: Any) -> ModuleOffsetVector:
        """Adds two vectors together, the other way."""
        if not isinstance(other, (LabwareOffsetVector, ModuleOffsetVector)):
            return NotImplemented
        return ModuleOffsetVector(
            x=other.x + self.x, y=other.y + self.y, z=other.z + self.z
        )

    def __sub__(self, other: Any) -> ModuleOffsetVector:
        """Subtracts two vectors."""
        if not isinstance(other, (LabwareOffsetVector, ModuleOffsetVector)):
            return NotImplemented
        return ModuleOffsetVector(
            x=self.x - other.x, y=self.y - other.y, z=self.z - other.z
        )

    def __rsub__(self, other: Any) -> ModuleOffsetVector:
        """Subtracts two vectors, the other way."""
        if not isinstance(other, (LabwareOffsetVector, ModuleOffsetVector)):
            return NotImplemented
        return ModuleOffsetVector(
            x=other.x - self.x, y=other.y - self.y, z=other.z - self.z
        )


@dataclass
class ModuleOffsetData:
    """Module calibration offset data."""

    moduleOffsetVector: ModuleOffsetVector
    location: DeckSlotLocation


class StackerFillEmptyStrategy(str, Enum):
    """Strategy to use for filling or emptying a stacker."""

    MANUAL_WITH_PAUSE = "manualWithPause"
    LOGICAL = "logical"



"""Protocol source value objects."""
from enum import Enum
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, Optional, Union
from typing_extensions import Literal

from opentrons.protocols.api_support.types import APIVersion

from opentrons_shared_data.robot.types import RobotType


class ProtocolType(str, Enum):
    """Type of protocol, JSON or Python."""

    JSON = "json"
    PYTHON = "python"


# TODO(mc, 2021-12-07): add python support roles
class ProtocolFileRole(str, Enum):
    """The purpose of a given file in a protocol.

    Args:
        MAIN: The protocol's main file. In a JSON protocol, this is will
            be the JSON file. In a Python protocol, this is the file
            that exports the main `run` method.
        LABWARE: A labware definition file, loadable by a
            Python file in the same protocol.
        DATA: An arbitrary text or csv file for usage in `bundled_data`
            for a Python protool
    """

    MAIN = "main"
    LABWARE = "labware"
    DATA = "data"


@dataclass(frozen=True)
class ProtocolSourceFile:
    """A single file in a protocol.

    Attributes:
        path: The file's path on disk.
        role: The file's purpose in the protocol.
    """

    path: Path
    role: ProtocolFileRole


@dataclass(frozen=True)
class JsonProtocolConfig:
    """Execution configuration for a JSON protocol.

    Attributes:
        schema_version: JSON schema version of the JSON protocol.
        protocol_type: Type of protocol
    """

    schema_version: int
    protocol_type: Literal[ProtocolType.JSON] = ProtocolType.JSON


@dataclass(frozen=True)
class PythonProtocolConfig:
    """Execution configuration for a Python protocol.

    Attributes:
        api_version: Python Protocol API version of the protocol.
        protocol_type: Type of protocol
    """

    api_version: APIVersion
    protocol_type: Literal[ProtocolType.PYTHON] = ProtocolType.PYTHON


ProtocolConfig = Union[JsonProtocolConfig, PythonProtocolConfig]
"""Union of all protocol execution configurations."""


# TODO(mc, 2021-12-09): Dict[str, Any] is an overly-permissive approximation
# due to mypy's lack of easy recursive types. Find a more accurate type
Metadata = Dict[str, Any]
"""A protocol's metadata (non-essential info, like author and title).

Robot software may not change how it executes a protocol based on
metadata (excepting a Python protocol's API version, which is in
metadata due to a historical implementation detail).

Metadata must be a simple JSON-serializable dictionary.
"""


@dataclass(frozen=True)
class ProtocolSource:
    """A value object representing a protocol and its source files on disk.

    This includes pointers to the files,
    plus some basic information that can be readily inferred from those files.
    (Excluding information that would require in-depth simulation of the protocol.)

    Attributes:
        directory: The directory containing the protocol files
            (and only the protocol files), or ``None`` if this is unknown.
        main_file: The location of the protocol's main file on disk.
        content_hash: A md5 hex digest of the protocol source combined with
            custom labware that was provided at the time of protocol creation
        files: Descriptions of all files that make up the protocol.
        metadata: Arbitrary metadata specified by the protocols.
        config: Protocol execution configuration.
        labware_definitions: Labware definitions provided by separate
            labware files or the main JSON protocol file, if present.
            This is not necessarily the same set of labware definitions
            that the protocol will actually attempt to load.
    """

    directory: Optional[Path]
    main_file: Path
    content_hash: str
    files: List[ProtocolSourceFile]
    metadata: Metadata
    robot_type: RobotType
    config: ProtocolConfig



"""A Protocol-Engine-friendly wrapper for opentrons.motion_planning.deck_conflict."""

from __future__ import annotations
import itertools
import logging
from typing import (
    Collection,
    Dict,
    Optional,
    Tuple,
    overload,
    Union,
    TYPE_CHECKING,
)

from opentrons_shared_data.errors.exceptions import MotionPlanningFailureError
from opentrons_shared_data.module import FLEX_TC_LID_COLLISION_ZONE

from opentrons.hardware_control.modules.types import ModuleType
from opentrons.motion_planning import deck_conflict as wrapped_deck_conflict

from opentrons.protocol_engine import (
    StateView,
    DeckSlotLocation,
    ModuleLocation,
    OnLabwareLocation,
    AddressableAreaLocation,
    InStackerHopperLocation,
    OFF_DECK_LOCATION,
    SYSTEM_LOCATION,
)
from opentrons.protocol_engine.errors.exceptions import LabwareNotLoadedOnModuleError
from opentrons.types import DeckSlotName, StagingSlotName, Point
from ...disposal_locations import TrashBin, WasteChute

if TYPE_CHECKING:
    from ...labware import Labware


class PartialTipMovementNotAllowedError(MotionPlanningFailureError):
    """Error raised when trying to perform a partial tip movement to an illegal location."""

    def __init__(self, message: str) -> None:
        super().__init__(
            message=message,
        )


class UnsuitableTiprackForPipetteMotion(MotionPlanningFailureError):
    """Error raised when trying to perform a pipette movement to a tip rack, based on adapter status."""

    def __init__(self, message: str) -> None:
        super().__init__(
            message=message,
        )


_log = logging.getLogger(__name__)

_FLEX_TC_LID_BACK_LEFT_PT = Point(
    x=FLEX_TC_LID_COLLISION_ZONE["back_left"]["x"],
    y=FLEX_TC_LID_COLLISION_ZONE["back_left"]["y"],
    z=FLEX_TC_LID_COLLISION_ZONE["back_left"]["z"],
)

_FLEX_TC_LID_FRONT_RIGHT_PT = Point(
    x=FLEX_TC_LID_COLLISION_ZONE["front_right"]["x"],
    y=FLEX_TC_LID_COLLISION_ZONE["front_right"]["y"],
    z=FLEX_TC_LID_COLLISION_ZONE["front_right"]["z"],
)


@overload
def check(
    *,
    engine_state: StateView,
    existing_labware_ids: Collection[str],
    existing_module_ids: Collection[str],
    existing_disposal_locations: Collection[Union[Labware, WasteChute, TrashBin]],
    new_labware_id: str,
) -> None:
    pass


@overload
def check(
    *,
    engine_state: StateView,
    existing_labware_ids: Collection[str],
    existing_module_ids: Collection[str],
    existing_disposal_locations: Collection[Union[Labware, WasteChute, TrashBin]],
    new_module_id: str,
) -> None:
    pass


@overload
def check(
    *,
    engine_state: StateView,
    existing_labware_ids: Collection[str],
    existing_module_ids: Collection[str],
    existing_disposal_locations: Collection[Union[Labware, WasteChute, TrashBin]],
    new_trash_bin: TrashBin,
) -> None:
    pass


def check(
    *,
    engine_state: StateView,
    existing_labware_ids: Collection[str],
    existing_module_ids: Collection[str],
    existing_disposal_locations: Collection[Union[Labware, WasteChute, TrashBin]],
    # TODO(mm, 2023-02-23): This interface is impossible to use correctly. In order
    # to have new_labware_id or new_module_id, the caller needs to have already loaded
    # the new item into Protocol Engine--but then, it's too late to do deck conflict.
    # checking. Find a way to do deck conflict checking before the new item is loaded.
    new_labware_id: Optional[str] = None,
    new_module_id: Optional[str] = None,
    new_trash_bin: Optional[TrashBin] = None,
) -> None:
    """Check for conflicts between items on the deck.

    This is a Protocol-Engine-friendly wrapper around
    opentrons.motion_planning.deck_conflict.check().

    Params:
        engine_state: An interface to retrieve details about the deck items.
        existing_labware_ids: The Protocol Engine IDs of all labware already loaded.
        existing_module_ids: The Protocol Engine IDs of all modules already loaded.
        new_labware_id: The Protocol Engine ID of a new labware you've just added.
            Mutually exclusive with new_module_id.
        new_module_id: The Protocol EngineID of a new module you've just added.
            Mutually exclusive with new_labware_id.

    Raises:
        opentrons.motion_planning.deck_conflict.DeckConflictError:
            If the newly-added item conflicts with one of the existing items.
    """

    if new_labware_id is not None:
        new_location_and_item = _map_labware(engine_state, new_labware_id)
    if new_module_id is not None:
        new_location_and_item = _map_module(engine_state, new_module_id)
    if new_trash_bin is not None:
        new_location_and_item = _map_disposal_location(new_trash_bin)

    if new_location_and_item is None:
        # The new item should be excluded from deck conflict checking. Nothing to do.
        return

    new_location, new_item = new_location_and_item

    all_existing_labware = (
        _map_labware(engine_state, labware_id) for labware_id in existing_labware_ids
    )
    mapped_existing_labware = (m for m in all_existing_labware if m is not None)

    all_existing_modules = (
        _map_module(engine_state, module_id) for module_id in existing_module_ids
    )
    mapped_existing_modules = (m for m in all_existing_modules if m is not None)

    all_exisiting_disposal_locations = (
        _map_disposal_location(disposal_location)
        for disposal_location in existing_disposal_locations
    )
    mapped_disposal_locations = (
        m for m in all_exisiting_disposal_locations if m is not None
    )

    existing_items: Dict[
        Union[DeckSlotName, StagingSlotName], wrapped_deck_conflict.DeckItem
    ] = {}
    for existing_location, existing_item in itertools.chain(
        mapped_existing_labware, mapped_existing_modules, mapped_disposal_locations
    ):
        assert existing_location not in existing_items
        existing_items[existing_location] = existing_item

    wrapped_deck_conflict.check(
        existing_items=existing_items,
        new_item=new_item,
        new_location=new_location,
        robot_type=engine_state.config.robot_type,
    )


def _map_labware(
    engine_state: StateView,
    labware_id: str,
) -> Optional[
    Tuple[Union[DeckSlotName, StagingSlotName], wrapped_deck_conflict.DeckItem]
]:
    location_from_engine = engine_state.labware.get_location(labware_id=labware_id)

    if isinstance(location_from_engine, AddressableAreaLocation):
        # This will be guaranteed to be either deck slot name or staging slot name
        slot: Union[DeckSlotName, StagingSlotName]
        try:
            slot = DeckSlotName.from_primitive(location_from_engine.addressableAreaName)
        except ValueError:
            slot = StagingSlotName.from_primitive(
                location_from_engine.addressableAreaName
            )
        return (
            slot,
            wrapped_deck_conflict.Labware(
                name_for_errors=engine_state.labware.get_load_name(
                    labware_id=labware_id
                ),
                highest_z=engine_state.geometry.get_labware_highest_z(
                    labware_id=labware_id
                ),
                uri=engine_state.labware.get_definition_uri(labware_id=labware_id),
                is_fixed_trash=engine_state.labware.is_fixed_trash(
                    labware_id=labware_id
                ),
            ),
        )

    elif isinstance(location_from_engine, DeckSlotLocation):
        # This labware is loaded directly into a deck slot.
        # Map it to a wrapped_deck_conflict.Labware.
        return (
            location_from_engine.slotName,
            wrapped_deck_conflict.Labware(
                name_for_errors=engine_state.labware.get_load_name(
                    labware_id=labware_id
                ),
                highest_z=engine_state.geometry.get_labware_highest_z(
                    labware_id=labware_id
                ),
                uri=engine_state.labware.get_definition_uri(labware_id=labware_id),
                is_fixed_trash=engine_state.labware.is_fixed_trash(
                    labware_id=labware_id
                ),
            ),
        )

    elif isinstance(location_from_engine, ModuleLocation):
        # This labware is loaded atop a module. Don't map it to anything here;
        # let _map_module() pick it up.
        return None

    elif isinstance(location_from_engine, OnLabwareLocation):
        # TODO(jbl 2023-06-08) check if we need to do any logic here or if this is correct
        return None

    elif (
        location_from_engine == OFF_DECK_LOCATION
        or location_from_engine == SYSTEM_LOCATION
        or isinstance(location_from_engine, InStackerHopperLocation)
    ):
        # This labware is off-deck. Exclude it from conflict checking.
        # todo(mm, 2023-02-23): Move this logic into wrapped_deck_conflict.
        return None


def _map_module(
    engine_state: StateView,
    module_id: str,
) -> Optional[Tuple[DeckSlotName, wrapped_deck_conflict.DeckItem]]:
    module_model = engine_state.modules.get_connected_model(module_id=module_id)
    module_type = module_model.as_type()
    mapped_location = engine_state.modules.get_location(module_id=module_id).slotName

    # Use the module model (e.g. "temperatureModuleV1") as the name for error messages
    # because it's convenient for us. Unfortunately, this won't necessarily match
    # the string that the Python protocol author used to load the module.
    name_for_errors = module_model.value

    highest_z_including_labware = _get_module_highest_z_including_labware(
        engine_state=engine_state,
        module_id=module_id,
    )

    if module_type == ModuleType.HEATER_SHAKER:
        return (
            mapped_location,
            wrapped_deck_conflict.HeaterShakerModule(
                name_for_errors=name_for_errors,
                highest_z_including_labware=highest_z_including_labware,
            ),
        )
    elif module_type == ModuleType.MAGNETIC_BLOCK:
        return (
            mapped_location,
            wrapped_deck_conflict.MagneticBlockModule(
                name_for_errors=name_for_errors,
                highest_z_including_labware=highest_z_including_labware,
            ),
        )
    elif module_type == ModuleType.THERMOCYCLER:
        return (
            mapped_location,
            wrapped_deck_conflict.ThermocyclerModule(
                name_for_errors=name_for_errors,
                highest_z_including_labware=highest_z_including_labware,
                # Python Protocol API >=v2.14 never allows loading a Thermocycler in
                # its semi configuration.
                is_semi_configuration=False,
            ),
        )
    elif module_type == ModuleType.FLEX_STACKER:
        # TODO: This is a placeholder. We need to implement this.
        return None
    else:
        return (
            mapped_location,
            wrapped_deck_conflict.OtherModule(
                name_for_errors=name_for_errors,
                highest_z_including_labware=highest_z_including_labware,
            ),
        )


def _map_disposal_location(
    disposal_location: Union[Labware, WasteChute, TrashBin],
) -> Optional[Tuple[DeckSlotName, wrapped_deck_conflict.DeckItem]]:
    if isinstance(disposal_location, TrashBin):
        return (
            disposal_location.location,
            wrapped_deck_conflict.TrashBin(
                name_for_errors="trash bin", highest_z=disposal_location.height
            ),
        )
    else:
        return None


def _deck_slot_to_int(deck_slot_location: DeckSlotLocation) -> int:
    return deck_slot_location.slotName.as_int()


def _get_module_highest_z_including_labware(
    engine_state: StateView, module_id: str
) -> float:
    try:
        labware_id = engine_state.labware.get_id_by_module(module_id=module_id)
    except LabwareNotLoadedOnModuleError:
        # No labware is loaded atop this module.
        # The height should be just the module itself.
        return engine_state.modules.get_overall_height(module_id=module_id)
    else:
        # This module has a labware loaded atop it. The height should include both.
        return engine_state.geometry.get_labware_highest_z(labware_id=labware_id)
